import os.path
import shelve
from .qt import QtGui, QtCore

import taxonome
from taxonome.config import config
from taxonome.taxa.name_selector import TryAnotherName
from .ui.prevchoices import Ui_PrevChoicesDialog
from .objects import NamePairItem

class PrevChoiceItem(QtGui.QStandardItem):
    def __init__(self, name, shelf):
        super().__init__()
        self.name = name
        self.shelf = shelf
        self.setText(str(name))
    
    @property
    def choice(self):
        return self.shelf[repr(self.name)]
    
    def update_choice(self, newchoice):
        options, oldchoice = self.choice
        self.shelf[repr(self.name)] = options, newchoice
        try:
            self.shelf.sync()
        except AttributeError:
            pass
    
    def reject_all(self):
        self.update_choice(KeyError(self.name))

class PrevChoicesDialog(QtGui.QDialog):
    def __init__(self, user_choices=None):
        QtGui.QDialog.__init__(self)
        self.ui = Ui_PrevChoicesDialog()
        self.ui.setupUi(self)
    
        if user_choices is not None:
            self.user_choices = user_choices
        else:
            user_choice_file = os.path.expanduser(config['cache']['user-choices'])
            self.user_choices = shelve.open(user_choice_file)
            self.destroyed.connect(self.user_choices.close)
        
        self.ambignames_model = QtGui.QStandardItemModel()
        self.ui.ambignames.setModel(self.ambignames_model)
        self.ui.ambignames.clicked.connect(self.select_ambigname)
        
        for ambigname in self.user_choices.keys():
            # Trivial bit of sanitisation - this is not secure against malicious
            # data in the user choices file.
            if not ambigname.startswith("taxonome.Name("):
                continue
            ambigname = eval(ambigname, {'taxonome':taxonome})
            self.ambignames_model.appendRow(PrevChoiceItem(ambigname, self.user_choices))
        
        # Connect action buttons
        self.ui.forget.clicked.connect(self.forget)
        self.ui.forget_all.clicked.connect(self.forget_all)
        self.ui.update_choice.clicked.connect(self.update_choice)
        self.ui.reject_all.clicked.connect(self.reject_all)
    
    @property
    def selected_choice(self):
        seln = self.ui.ambignames.selectedIndexes()
        return self.ambignames_model.itemFromIndex(seln[0])
    
    @property
    def selected_name(self):
        seln = self.ui.altmatches.selectedIndexes()
        return self.options_model.itemFromIndex(seln[0])
        
    def select_ambigname(self, index):
        prevchoice = self.ambignames_model.itemFromIndex(index)
        self.show_choices(prevchoice)
    
    def show_choices(self, prevchoice):
        options, chosen = prevchoice.choice
        
        if isinstance(chosen, KeyError):
            chosentxt = "(Reject all)"
        elif isinstance(chosen, TryAnotherName):
            chosentxt = "Retry with '{}'".format(chosen.newname)
        else:
            n, an = chosen
            if n==an:
                chosentxt = str(n)
            else:
                chosentxt = "{an}<br>&nbsp;&nbsp;from {n}".format(n=n, an=an)
        self.ui.chosenmatch.setText(chosentxt)
        
        self.options_model = QtGui.QStandardItemModel()
        for namepair in options:
            self.options_model.appendRow(NamePairItem(namepair))
        self.ui.altmatches.setModel(self.options_model)
    
    def forget(self):
        choice = self.selected_choice
        del self.user_choices[repr(choice.name)]
        row = choice.index().row()
        self.ambignames_model.removeRow(row)
    
    def forget_all(self):
        self.user_choices.clear()
        self.ambignames_model = QtGui.QStandardItemModel()
        self.ui.ambignames.setModel(self.ambignames_model)
    
    def update_choice(self):
        choice = self.selected_choice
        name = self.selected_name
        choice.update_choice(name.namepair)
        self.show_choices(choice)
    
    def reject_all(self):
        choice = self.selected_choice
        choice.reject_all()
        self.show_choices(choice)

def edit_prevchoices(app=None):
    user_choices = None
    if app is not None:
        user_choices = app.gui_select_name.previous_choices
    dialog = PrevChoicesDialog(user_choices)
    dialog.exec_()



from .qt import QtGui

from taxonome.services import grin, tropicos, col
from .ui.fetch_taxa import Ui_FetchOptsDialog
from .ui.find_taxon import Ui_FindSpDialog

from . import iothread


class WrappedWoRMS:
    """Lazy loading so we don't import suds at startup."""
    __worms_resource = None
    
    @property
    def worms_resource(self):
        from taxonome.services.worms import WoRMSTaxaResource
        self.__worms_resource = WoRMSTaxaResource()
        return self.__worms_resource
    
    def select(self, *args, **kwargs):
        return self.worms_resource.select(*args, **kwargs)

fetch_taxa_services = {"USDA GRIN database": grin}
find_taxon_services = {"USDA GRIN database": grin,
                       "MBG Tropicos": tropicos,
                       "Catalogue of Life": col,
                       "World Register of Marine Species (WoRMS)": WrappedWoRMS(),
                      }

def fetch_taxa(app):
    optsdialog = Ui_FetchOptsDialog()
    dialog = QtGui.QDialog()
    optsdialog.setupUi(dialog)
    
    optsdialog.service.addItems(list(fetch_taxa_services.keys()))
    
    if not dialog.exec_():
        return False
    
    service_name = optsdialog.service.currentText()
    service = fetch_taxa_services[service_name]
    name = optsdialog.groupname.text()
    
    ds_name = "{} from {}".format(name, service_name)
    app.ui.statusbar.showMessage("Fetching {}...".format(ds_name))
    iothread.makeloader(app, lambda: service.fetch(name), ds_name)
    #ts = service.fetch(name)
    #app.add_dataset(, ts)

def find_taxon(app):
    tax = None
    finddialog = Ui_FindSpDialog()
    dialog = QtGui.QDialog()
    finddialog.setupUi(dialog)
    
    finddialog.service.addItems(list(find_taxon_services.keys()))
    finddialog.destination_ds.setModel(app.datasets_model)
    
    def dosearch():
        nonlocal tax
        service_name = finddialog.service.currentText()
        service = find_taxon_services[service_name]
        name = finddialog.taxname.text()
        
        tax = service.select(name)
        
        finddialog.result_view.setHtml(tax.html())
        finddialog.result_block.setEnabled(True)
    
    def add_to_ds():
        seln = finddialog.destination_ds.currentIndex()
        target = app.datasets_model.item(seln)
        
        target.ds.add(tax)
        target.emitDataChanged()
    
    finddialog.search.clicked.connect(dosearch)
    finddialog.ds_add.clicked.connect(add_to_ds)
    
    dialog.exec_()
    



def htmlise(obj):
    if isinstance(obj, dict):
        return dict_html(obj)
    if isinstance(obj, (list, set, tuple)):
        return list_html(obj)
    return str(obj)

_dict_html_template = "<ul>{}</ul>"
_item_html_template = "<li><strong>{}:</strong> {}</li>"

def dict_html(d):
    items = (_item_html_template.format(str(k), htmlise(v)) for k,v in d.items())
    return _dict_html_template.format("\n".join(items))
        
def list_html(l):
    return ", ".join(l)



import unittest
from .. import regions
from taxonome.utils.testing import check_func

class RegionsTest(unittest.TestCase):
    def testTDWG(self):
        if not regions.world:
            regions.load_tdwg()
        world = regions.world
        assert len(regions.tdwg_levels[1]) == 9
        assert world.is_subregion(regions.tdwg["4"], regions.tdwg["EHM-BH"])
        assert not world.is_subregion(regions.tdwg["3"], regions.names["Bhutan"])
        assert world.is_subregion(regions.ISO["GB"], regions.tdwg["IRE-NI"])
        assert not world.is_subregion(regions.tdwg["26"], regions.names["Netherlands"])
        assert len(list(world.subregions(regions.tdwg["36"]))) == 44
        assert len(world.immediate_subregions(regions.tdwg["36"])) == 9
    
    def test_find_tdwg(self):
        tests = [("Netherlands", {"NET"}),
                 ("Ecuador", {"ECU"}),
                 ("Chongqing", {"CHC"}),   # Level 4 -> 3
                 ("Brazil", {'BZS', 'BZE', 'BZC', 'BZN', 'BZL'}), # Level 2 -> 3
                 ("Turkey", {'TUE', 'TUR'}) # TUE is 'Turkey-in-Europe'
                ]
        check_func(regions.find_tdwg, tests)
        
        with self.assertRaises(KeyError):
            regions.find_tdwg("Mars")
        
        self.assertEqual(regions.find_tdwg("OM",level=2, index=regions.ISO), {"35"})
    
    def test_tdwgise(self):
        """tdwgise should break down larger regions (Brazil), and find parent
        regions (Chongqing)."""
        raw = ["Netherlands","Ecuador", "Brazil", "Chongqing", "Mars", "Turkey"]
        exp = {'CHC', 'BZS', 'ECU', 'BZE', 'BZC', 'BZN', 'NET', 'BZL', 'TUR', 'TUE'}
        tdwged, notfound = regions.tdwgise(raw)
        self.assertEqual(tdwged, exp)
        self.assertEqual(notfound, {"Mars"})
        
if __name__ == "__main__":
    unittest.main()



import builtins
import tempfile
import unittest
from os import remove

import taxonome
from taxonome.taxa import file_csv, name_selector
from taxonome.taxa.base import UncertainSpeciesError
from taxonome.taxa.author import Authority
import taxonome.taxa.collection

Name = taxonome.Name

class AutoInput:
    """Context manager to override the input() call. Will automatically
    return the given response."""
    def __init__(self, response=''):
        self.response = response
    
    def __enter__(self):
        self.saved_input = builtins.input
        def autoinput(prompt):
            return self.response
        builtins.input = autoinput
    
    def __exit__(self, exc_type, exc_value, traceback):
        builtins.input = self.saved_input

class HyacinthBeanTest(unittest.TestCase):
    def setUp(self):
        self.bean = taxonome.Taxon("Lablab purpureus", "(L.) Sweet")
        self.bean.othernames.add(Name("Dolichos lablab","L."))
        self.bean.othernames.add(Name("Lablab niger", "Medikus"))
        self.bean.othernames.add(Name("Vigna aristata", "Piper"))
        self.ts = taxonome.TaxonSet()
        self.ts.add(self.bean)
    
    def testSynonymy(self):
        assert self.bean.hasname("Dolichos lablab")
        assert self.bean.hasname(Name("Lablab niger", "Medikus"))
    
    def testShortauth(self):
        assert self.bean.name == Name("Lablab purpureus", "Sweet.")
        assert self.bean.name == Name("Lablab purpureus", "Sweet ex. Madeup")
        
    def testTSSynonymy(self):
        assert Name("Vigna aristata","Piper.") in self.ts
        assert self.ts["Lablab niger"]
    
    def testAccnameLookup(self):
        self.assertEqual(self.ts.get_by_accepted_name(Name("Lablab purpureus",
                                "(L.) Sweet")), self.bean)
        with self.assertRaises(KeyError):
            self.ts.get_by_accepted_name(Name("Dolichos lablab","L."))
    
    def test_non_auth_lookup(self):
        t1 = self.ts.select(Name("Dolichos lablab", "L."))
        self.assertIs(t1, self.bean)
        with self.assertRaises(KeyError):
            self.ts.select(Name("Dolichos lablab", "non L."))
        with self.assertRaises(KeyError):
            self.ts.select(Name("Dolichos lablab", "sensu non L."))
        with self.assertRaises(KeyError):
            self.ts.select(Name("Dolichos lablab", "auct. non L."))
        
        # Test selecting the other way round.
        ts2 = taxonome.TaxonSet()
        ts2.add(taxonome.Taxon("Dolichos lablab", "non L."))
        with self.assertRaises(KeyError):
            ts2.select(Name("Dolichos lablab", "L."))
        
        # This should succeed.
        ts2.select(Name("Dolichos lablab", "Vahl"), strict_authority=False)

class PoaTest(unittest.TestCase):
    def setUp(self):
        self.t1 = taxonome.Taxon("Poa annua", "L.")
        self.t2 = taxonome.Taxon("Poa infirma", "H. B. & K.")
        self.t2.othernames.add(Name("Poa annua", "Cham. & Schlecht."))
        self.ts = taxonome.TaxonSet()
        self.ts.add(self.t1)
        self.ts.add(self.t2)
        self.ts._nameselector = name_selector.TerminalNameSelector(previous_choices={})
    
    def testPreferAccName(self):
        # No authority: should pick accepted name without asking for input.
        with AutoInput("N"):
            t1 = self.ts.select("Poa annua")
        self.assertEqual(t1, self.t1)
        
        # With authority: should ask the user by default.
        with AutoInput("0"), self.assertRaisesRegex(KeyError, "0"):
            self.ts.select(Name("Poa annua", "Madeup"))
        
        # With authority: When requested, prefer the accepted name:
        with AutoInput("N"):
            t1a = self.ts.select(Name("Poa annua", "Madeup"), prefer_accepted='all')
        self.assertEqual(t1a, self.t1)
        
class LiquoriceTestBase(unittest.TestCase):
    def setUp(self):
        self.ts = taxonome.TaxonSet()
        self.liquorice = taxonome.Taxon("Glycyrrhiza glabra", "L.")
        self.liquorice.othernames.add(Name("Glycyrrhiza glandulifera", "Waldst. and Kit."))
        self.ts.add(self.liquorice)
        ch_liquorice = taxonome.Taxon("Glycyrrhiza uralensis", "Fisch.")
        ch_liquorice.othernames.add(Name("Glycyrrhiza glandulifera", "Ledeb."))
        self.ts.add(ch_liquorice)

class LiquoriceTest(LiquoriceTestBase):
    def testHomonym(self):
        assert self.ts[Name("Glycyrrhiza glabra", "L.")].hasname(Name("Glycyrrhiza glandulifera", "Waldst. & Kit."))
        homonyms = self.ts.resolve_name("Glycyrrhiza glandulifera")
        assert len(homonyms) == 2
        assert homonyms[0][0].match_unqualified(homonyms[1][0])
        assert not homonyms[0][1].match_unqualified(homonyms[1][1])
        with AutoInput("1"):
            self.assertIn(self.ts.select("Glycyrrhiza glandulifera"), self.ts)
    
    def testFuzzyLookup(self):
        fuzzy_match = self.ts.select("Glycyrhiza glabra")
        self.assertEqual(self.liquorice, fuzzy_match)
        with self.assertRaises(KeyError):
            self.ts.select("Glycyrrhiza echinata")
        
    def testUpgradeSubsp(self):
        n1 = Name("Glycyrrhiza glabra subsp. glabra")
        n2 = Name("Glycyrrhiza glabra var. glabra")
        n3 = Name("Glycyrrhiza glabra subsp. spammens")
        with self.assertRaises(KeyError):
            self.ts.select(n1, upgrade_subsp="none")
        match1 = self.ts.select(n1, upgrade_subsp="nominal")
        self.assertEqual(match1, self.liquorice)
        match2 = self.ts.select(n2, upgrade_subsp="nominal")
        self.assertEqual(match2, self.liquorice)
        with self.assertRaises(KeyError):
            self.ts.select(n3, upgrade_subsp="nominal")
        match3 = self.ts.select(n3, upgrade_subsp="all")
        self.assertEqual(match3, self.liquorice)
        
    def testDeletionByString(self):
        with self.assertRaises(KeyError):
            del self.ts["Glycyrrhiza glandulifera"]
        del self.ts["Glycyrrhiza glabra"]
        assert taxonome.Name("Glycyrrhiza glabra", "L.") not in self.ts
        assert len(self.ts.resolve("Glycyrrhiza uralensis")) == 1
        
    def testDeletionByQName(self):
        name = taxonome.Name("Glycyrrhiza glandulifera", "Ledeb.")
        del self.ts[name]
        assert name not in self.ts
        assert "Glycyrrhiza uralensis" not in self.ts

class NameParseTest(unittest.TestCase):
    def test_parse_name(self):
        n = Name.from_string("Teramnus repens (Taub.)Baker f. ssp. gracilis (Chiov.)Verdc.")
        self.assertEqual(n.plain, "Teramnus repens subsp. gracilis")
        self.assertEqual(str(n.authority), "(Chiov.) Verdc.")
        parent = n.parent
        self.assertEqual(parent.plain, "Teramnus repens")
        # f. (filius, i.e. son) is ignored for now
        self.assertEqual(str(parent.authority), "(Taub.) Baker")
        
        # N.B. var without a .
        n = Name("Vigna radiata var sublobata", "(Roxb.) Verdc.")
        self.assertEqual(n.plain, "Vigna radiata var. sublobata")
        
        # Extra spaces
        n = Name("Vigna    radiata\tvar. \t sublobata")
        self.assertEqual(n.plain, "Vigna radiata var. sublobata")
    
    def test_subspecies(self):
        n = Name("Gorilla gorilla gorilla", "Savage, 1847")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Gorilla gorilla")
        
        n = Name.from_string("Gorilla gorilla gorilla Savage, 1847")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Gorilla gorilla")
        
        n = Name("Alloteropsis semialata subsp. eckloniana", "(Nees) Gibbs-Russ.")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Alloteropsis semialata")
        
        n = Name.from_string("Alloteropsis semialata (R. Br.) Hitchc. subsp. eckloniana (Nees) Gibbs-Russ.")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Alloteropsis semialata")
        self.assertEqual(str(n.parent.authority), "(R. Br.) Hitchc.")
        self.assertEqual(str(n.authority), "(Nees) Gibbs-Russ.")
        
    def test_forma(self):
        n = Name('Muhlenbergia mexicana f. ambigua')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        
        n = Name.from_string('Muhlenbergia mexicana L. f. ambigua Foo.')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        self.assertEqual(str(n.authority), "Foo.")
        self.assertEqual(str(n.parent.authority), "L.")
        
        n = Name('Muhlenbergia mexicana forma mexicana')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        
        n = Name.from_string('Muhlenbergia mexicana L. forma mexicana Bar.')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        self.assertEqual(str(n.authority), "Bar.")
        self.assertEqual(str(n.parent.authority), "L.")
        
        # With f. for filius
        n = Name.from_string('Crotalaria dolichonyx Baker f. & Martin')
        self.assertEqual(n.rank, 'species')
        self.assertEqual(n.plain, 'Crotalaria dolichonyx')
        self.assertEqual(n.authority, Authority('Baker f. & Martin'))
    
    def test_hybrid(self):
        names = [('X Achnella caduca', 'Achnella caduca', 'genus'),
                 ('x Achnella caduca', 'Achnella caduca', 'genus'),
                 ('× Achnella caduca', 'Achnella caduca', 'genus'),
                 ('Calammophila X baltica', 'Calammophila baltica', 'species'),
                 ('Calammophila x baltica', 'Calammophila baltica', 'species'),
                 ('Calammophila × baltica', 'Calammophila baltica', 'species'),
                ]
        
        for raw, plain, hybrid in names:
            print(raw)
            n = Name(raw)
            self.assertEqual(n.plain, plain)
            self.assertEqual(n.hybrid, hybrid)
            
            n = Name.from_string(raw + " Bloggs")
            self.assertEqual(n.plain, plain)
            self.assertEqual(str(n.authority), "Bloggs")
            self.assertEqual(n.hybrid, hybrid)
        
    def test_genus_name(self):
        n = Name("Alloteropsis", "Presl.")
        self.assertEqual(n.plain, "Alloteropsis")
        self.assertEqual(str(n.authority), "Presl.")
    
    def test_with_year(self):
        n = Name.from_string("Balaena mysticetus Linnaeus, 1758")
        self.assertEqual(n.plain, "Balaena mysticetus")
        self.assertEqual(n.authority.year, "1758")
        
        self.assertEqual(n, Name("Balaena mysticetus", "L."))
    
    def test_invalid_name(self):
        with self.assertRaises(UncertainSpeciesError):
            Name.from_string("Festuca ovina & rubra")
        
        with self.assertRaises(UncertainSpeciesError):
            Name.from_string("Festuca sp.")
        with self.assertRaises(UncertainSpeciesError):
            Name.from_string("Festuca spp.")
        
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria sp")
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria sp.")
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria spp")
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria spp.")
    
    def test_not_capitalised(self):
        n = Name.from_string("cynodon dactylon")
        self.assertEqual(n.plain, "Cynodon dactylon")
        n = Name("echinochloa haploclada", "")
        self.assertEqual(n.plain, "Echinochloa haploclada")
    
    def test_auth_noncapital(self):
        n = Name.from_string("Trifolium pauciflorum d'Urv.")
        self.assertEqual(n.plain, "Trifolium pauciflorum")
        self.assertEqual(str(n.authority), "d'Urv.")
        
        n = Name.from_string("Trifolium pauciflorum de Madeup")
        self.assertEqual(n.plain, "Trifolium pauciflorum")
        self.assertEqual(str(n.authority), "de Madeup")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'taxonome/gui/ui/fetch_taxa.ui'
#
# Created: Tue May 29 11:46:18 2012
#      by: PyQt4 UI code generator 4.8.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class Ui_FetchOptsDialog(object):
    def setupUi(self, FetchOptsDialog):
        FetchOptsDialog.setObjectName(_fromUtf8("FetchOptsDialog"))
        FetchOptsDialog.resize(400, 300)
        self.verticalLayout = QtGui.QVBoxLayout(FetchOptsDialog)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.label = QtGui.QLabel(FetchOptsDialog)
        self.label.setObjectName(_fromUtf8("label"))
        self.horizontalLayout.addWidget(self.label)
        spacerItem = QtGui.QSpacerItem(20, 20, QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.groupname = QtGui.QLineEdit(FetchOptsDialog)
        self.groupname.setObjectName(_fromUtf8("groupname"))
        self.horizontalLayout.addWidget(self.groupname)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.label_2 = QtGui.QLabel(FetchOptsDialog)
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.horizontalLayout_2.addWidget(self.label_2)
        self.service = QtGui.QComboBox(FetchOptsDialog)
        self.service.setObjectName(_fromUtf8("service"))
        self.horizontalLayout_2.addWidget(self.service)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.label_3 = QtGui.QLabel(FetchOptsDialog)
        self.label_3.setWordWrap(True)
        self.label_3.setObjectName(_fromUtf8("label_3"))
        self.verticalLayout.addWidget(self.label_3)
        self.buttonBox = QtGui.QDialogButtonBox(FetchOptsDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
        self.verticalLayout.addWidget(self.buttonBox)
        self.label.setBuddy(self.groupname)
        self.label_2.setBuddy(self.service)

        self.retranslateUi(FetchOptsDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), FetchOptsDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), FetchOptsDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(FetchOptsDialog)

    def retranslateUi(self, FetchOptsDialog):
        FetchOptsDialog.setWindowTitle(QtGui.QApplication.translate("FetchOptsDialog", "Dialog", None, QtGui.QApplication.UnicodeUTF8))
        self.label.setText(QtGui.QApplication.translate("FetchOptsDialog", "Group name", None, QtGui.QApplication.UnicodeUTF8))
        self.label_2.setText(QtGui.QApplication.translate("FetchOptsDialog", "Source", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setText(QtGui.QApplication.translate("FetchOptsDialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Ubuntu\'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">This will download basic data for species in  a taxonomic group, e.g. <span style=\" font-style:italic;\">Poaceae</span>. So far, this only works with GRIN, where it can access synonymy data.</p></body></html>", None, QtGui.QApplication.UnicodeUTF8))




# -*- coding: utf-8 -*-

"""The user interface for our app"""

import csv
import os.path
import traceback
import sys
import warnings
from base64 import b64encode

from .qt import QtCore, QtGui, QtWebKit

from .ui.mainwindow import Ui_MainWindow
from . import (selectname, mapnames, combine, webservices, iothread, csvdialogs,
               prevchoices)
from .objects import DatasetItem, TaxonItem, TaxaItemModel

from taxonome import TaxonSet
from taxonome.taxa import name_selector
from taxonome.taxa import file_csv, file_jsonlines

# Create a class for our main window
class Main(QtGui.QMainWindow):
    
    active_dataset = None
    
    def __init__(self):
        QtGui.QMainWindow.__init__(self)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        
        # Keep a reference to worker threads so they don't get destroyed
        # prematurely.
        self.workerthreads = set()
        self.warning.connect(self.showwarning)
        
        self.lastdir = os.path.expanduser("~")
        
        # List of datasets
        self.datasets_model = QtGui.QStandardItemModel()
        self.ui.datasets.setModel(self.datasets_model)
        self.ui.datasets.clicked.connect(self.select_dataset)
        self.datasets_model.itemChanged.connect(DatasetItem.update_name)
        
        # Preview pane
        blank_preview = TaxaItemModel(self)
        self.ui.taxa_list.setModel(blank_preview)
        self.ui.taxa_list.setColumnWidth(0, 200)
        self.ui.taxa_list.setColumnWidth(1, 100)
        self.ui.taxa_list.clicked.connect(self.select_taxon)
        
        # HTML taxon preview
        websettings = QtWebKit.QWebSettings.globalSettings()
        styles = b"body {font-family: sans-serif;}"
        stylesheet = "data:text/css;charset=utf-8;base64," + b64encode(styles).decode('ascii')
        websettings.setUserStyleSheetUrl(QtCore.QUrl(stylesheet))
        
        # Set up menus
        #  File:
        self.ui.actionNew_ds.triggered.connect(self.empty_ds)
        self.ui.actionLoad_CSV_taxa.triggered.connect(self.load_csv)
        self.ui.actionLoad_CSV_synonyms.triggered.connect(self.load_csv_synonyms)
        self.ui.actionLoad_CSV_individuals.triggered.connect(self.load_csv_individuals)
        self.ui.actionLoad_JSONlines.triggered.connect(self.load_jsonlines)
        self.ui.actionSave_CSV_taxa.triggered.connect(self.save_csv)
        self.ui.actionSave_CSV_synonyms.triggered.connect(self.save_csv_synonyms)
        self.ui.actionSave_CSV_individuals.triggered.connect(self.save_csv_individuals)
        self.ui.actionSave_JSONlines.triggered.connect(self.save_jsonlines)
        #  Taxa:
        self.ui.actionMatch_names.triggered.connect(self.map_names)
        self.ui.actionCombine_datasets.triggered.connect(self.combine_datasets)
        self.ui.actionPrevious_choices.triggered.connect(self.edit_prevchoices)
        #  Web services:
        self.ui.actionFetch_taxa.triggered.connect(self.fetch_taxa)
        self.ui.actionLook_up_name.triggered.connect(self.find_taxon_web)
        
        # Status bar
        self.ui.statusbar.showMessage("Welcome to Taxonome.", 10000)
        
        self.gui_select_name = selectname.GuiNameSelector()
        self.gui_select_name.choice_required.connect(self.gui_select_name.present_dialog)
        self.silent_select_name = name_selector.NameSelector()
    
    def select_dataset(self, index):
        ds_item = self.datasets_model.itemFromIndex(index)
        self.active_dataset = ds_item
        if ds_item.search_results:
            self.ui.taxa_list.setModel(ds_item.search_results)
            self.ui.searchbox.setText(ds_item.search_term)
            self.ui.taxa_list_label.setText("Search results:")
        else:
            self.ui.taxa_list.setModel(ds_item.preview)
            self.ui.searchbox.clear()
            self.ui.taxa_list_label.setText("Example taxa:")
    
    def select_taxon(self, index):
        tax_item = self.ui.taxa_list.model().itemFromIndex(index)
        self.ui.tax_preview.setHtml(tax_item.taxon.html())
    
    def add_dataset(self, name, ts, filename=None):
        ds_item = DatasetItem(name, ts, filename=filename)
        self.datasets_model.appendRow(ds_item)
        
        ds_item.preview = TaxaItemModel(self)
        its = iter(ts)
        for _ in range(10):
            try:
                tax = next(its)
            except StopIteration:
                break
            name = TaxonItem(tax.name.plain, taxon=tax)
            auth = TaxonItem(str(tax.name.authority), taxon=tax)
            ds_item.preview.appendRow([name, auth])
    
    def empty_ds(self):
        ts = TaxonSet()
        self.add_dataset("unnamed", ts)
    
    def load_jsonlines(self):
        filename = QtGui.QFileDialog.getOpenFileName(self, "Open file",
                    self.lastdir, "JSONlines (*.jsonlines);;All files (*.*)")
        if not filename:
            return
        self.lastdir = os.path.dirname(filename)
        
        ds_name = os.path.splitext(os.path.basename(filename))[0].replace("_"," ")
        
        iothread.load_jsonlines(self, ds_name, filename)
    
    def save_jsonlines(self):
        seln = self.ui.datasets.selectedIndexes()
        if not seln:
            return
        ds_item = self.datasets_model.itemFromIndex(seln[0])
        filename = os.path.splitext(ds_item.autofilename)[0] + ".jsonlines"
        filename = QtGui.QFileDialog.getSaveFileName(self, "Save file", filename,
                                "JSONlines (*.jsonlines);;All files (*.*)")
        if not filename:
            return
        if not os.path.splitext(filename)[1]:
            filename += ".jsonlines"
        
        with open(filename,"w", encoding='utf-8') as f:
            file_jsonlines.save_taxa(f, ds_item.ds)
    
    def save_csv(self):
        ds_item = self.active_dataset
        if not ds_item:
            return
        filename = os.path.splitext(ds_item.autofilename)[0] + ".csv"
        filename = QtGui.QFileDialog.getSaveFileName(self, "Save file", filename,
                                                "CSV (*.csv);;All files (*.*)")
        if not filename:
            return
        if not os.path.splitext(filename)[1]:
            filename += ".csv"
        
        with open(filename,"w",encoding='utf-8', newline='') as f:
            file_csv.save_taxa(f, ds_item.ds, info_fields=True, write_distribution=None)
    
    def save_csv_synonyms(self):
        ds_item = self.active_dataset
        if not ds_item:
            return
        filename = os.path.splitext(ds_item.autofilename)[0] + ".csv"
        filename = QtGui.QFileDialog.getSaveFileName(self, "Save file", filename,
                                                "CSV (*.csv);;All files (*.*)")
        if not filename:
            return
        if not os.path.splitext(filename)[1]:
            filename += ".csv"
        
        with open(filename,"w",encoding='utf-8', newline='') as f:
            file_csv.save_synonyms(f, ds_item.ds)
    
    def save_csv_individuals(self):
        ds_item = self.active_dataset
        if not ds_item:
            return
        filename = os.path.splitext(ds_item.autofilename)[0] + ".csv"
        filename = QtGui.QFileDialog.getSaveFileName(self, "Save file", filename,
                                                "CSV (*.csv);;All files (*.*)")
        if not filename:
            return
        if not os.path.splitext(filename)[1]:
            filename += ".csv"
        
        with open(filename,"w",encoding='utf-8', newline='') as f:
            file_csv.save_individuals(f, ds_item.ds)
    
    load_csv = csvdialogs.load_csv
    load_csv_synonyms = csvdialogs.load_csv_synonyms
    load_csv_individuals = csvdialogs.load_csv_individuals
    
    def do_search(self):
        """Search for a taxon name in the active dataset."""
        term = self.ui.searchbox.text()
        wildcard = "*" in term
        try:
            results = self.active_dataset.ds.resolve(term, wildcard=wildcard)
        except NotImplementedError:
            self.ui.statusbar.showMessage("This data source doesn't support "
                                          "wildcard searches.")
            return
        results_model = TaxaItemModel(self)
        for tax in results:
            name = TaxonItem(tax.name.plain, taxon=tax)
            auth = TaxonItem(str(tax.name.authority), taxon=tax)
            results_model.appendRow([name, auth])
        
        # Display the results
        self.ui.taxa_list.setModel(results_model)
        self.ui.taxa_list_label.setText("Search results:")
        
        # Store results with the dataset for later use.
        self.active_dataset.search_results = results_model
        self.active_dataset.search_term = term
    
    def clear_search(self):
        """Clear search results. The input field is cleared separately by
        a connection to its own slot."""
        self.ui.taxa_list.setModel(self.active_dataset.preview)
        self.ui.taxa_list_label.setText("Example taxa:")
        
        self.active_dataset.search_results = None
        self.active_dataset.search_term = None
    
    def show_error(self, e):
        self.excepthook(type(e), e, e.__traceback__)
    
    def excepthook(self, etype, value, tb):
        msgbox = QtGui.QMessageBox(self)
        err = "".join(traceback.format_exception(etype, value, tb))
        msgbox.setText("An exception occurred:\n" + err +\
                       "\n\nPlease report this at https://bitbucket.org/taxonome/taxonome/issues")
        msgbox.exec_()
    
    warning = QtCore.Signal(object, object, object, object, object, object)
    def thread_warn(self, message, category, filename, lineno, file=None, line=None):
        self.warning.emit(message, category, filename, lineno, file, line)
    
    def showwarning(self, message, category, filename, lineno, file=None, line=None):
        self.ui.statusbar.showMessage("Warning: " + str(message), 10000)
    
    map_names = mapnames.wizard_map_names
    combine_datasets = combine.gui_combine_datasets
    fetch_taxa = webservices.fetch_taxa
    find_taxon_web = webservices.find_taxon
    edit_prevchoices = prevchoices.edit_prevchoices
        
def main():
    app = QtGui.QApplication(sys.argv)
    window = Main()
    warnings.showwarning = window.thread_warn
    if sys.stderr is None:
        sys.excepthook = window.excepthook
    window.show()
    sys.exit(app.exec_())



try:
    from PyQt4 import QtCore, QtGui, QtWebKit, QtNetwork
    QtCore.Signal = QtCore.pyqtSignal
    QtCore.Slot = QtCore.pyqtSlot
except ImportError:
    from PySide import QtCore, QtGui, QtWebKit, QtNetwork



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'taxonome/gui/ui/find_taxon.ui'
#
# Created: Tue May 29 11:46:18 2012
#      by: PyQt4 UI code generator 4.8.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class Ui_FindSpDialog(object):
    def setupUi(self, FindSpDialog):
        FindSpDialog.setObjectName(_fromUtf8("FindSpDialog"))
        FindSpDialog.resize(380, 354)
        self.verticalLayout = QtGui.QVBoxLayout(FindSpDialog)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.formLayout_2 = QtGui.QFormLayout()
        self.formLayout_2.setFieldGrowthPolicy(QtGui.QFormLayout.AllNonFixedFieldsGrow)
        self.formLayout_2.setObjectName(_fromUtf8("formLayout_2"))
        self.label = QtGui.QLabel(FindSpDialog)
        self.label.setObjectName(_fromUtf8("label"))
        self.formLayout_2.setWidget(1, QtGui.QFormLayout.LabelRole, self.label)
        self.label_2 = QtGui.QLabel(FindSpDialog)
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.formLayout_2.setWidget(3, QtGui.QFormLayout.LabelRole, self.label_2)
        self.service = QtGui.QComboBox(FindSpDialog)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.service.sizePolicy().hasHeightForWidth())
        self.service.setSizePolicy(sizePolicy)
        self.service.setObjectName(_fromUtf8("service"))
        self.formLayout_2.setWidget(3, QtGui.QFormLayout.FieldRole, self.service)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.taxname = QtGui.QLineEdit(FindSpDialog)
        self.taxname.setObjectName(_fromUtf8("taxname"))
        self.horizontalLayout_2.addWidget(self.taxname)
        self.search = QtGui.QPushButton(FindSpDialog)
        self.search.setObjectName(_fromUtf8("search"))
        self.horizontalLayout_2.addWidget(self.search)
        self.formLayout_2.setLayout(1, QtGui.QFormLayout.FieldRole, self.horizontalLayout_2)
        self.verticalLayout.addLayout(self.formLayout_2)
        self.result_block = QtGui.QWidget(FindSpDialog)
        self.result_block.setEnabled(False)
        self.result_block.setObjectName(_fromUtf8("result_block"))
        self.verticalLayout_2 = QtGui.QVBoxLayout(self.result_block)
        self.verticalLayout_2.setMargin(0)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.result_view = QtWebKit.QWebView(self.result_block)
        self.result_view.setUrl(QtCore.QUrl(_fromUtf8("about:blank")))
        self.result_view.setObjectName(_fromUtf8("result_view"))
        self.verticalLayout_2.addWidget(self.result_view)
        self.horizontalLayout_3 = QtGui.QHBoxLayout()
        self.horizontalLayout_3.setObjectName(_fromUtf8("horizontalLayout_3"))
        self.label_3 = QtGui.QLabel(self.result_block)
        self.label_3.setObjectName(_fromUtf8("label_3"))
        self.horizontalLayout_3.addWidget(self.label_3)
        self.destination_ds = QtGui.QComboBox(self.result_block)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.destination_ds.sizePolicy().hasHeightForWidth())
        self.destination_ds.setSizePolicy(sizePolicy)
        self.destination_ds.setObjectName(_fromUtf8("destination_ds"))
        self.horizontalLayout_3.addWidget(self.destination_ds)
        self.ds_add = QtGui.QPushButton(self.result_block)
        self.ds_add.setObjectName(_fromUtf8("ds_add"))
        self.horizontalLayout_3.addWidget(self.ds_add)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.verticalLayout.addWidget(self.result_block)
        self.buttonBox = QtGui.QDialogButtonBox(FindSpDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Close)
        self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(FindSpDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), FindSpDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), FindSpDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(FindSpDialog)

    def retranslateUi(self, FindSpDialog):
        FindSpDialog.setWindowTitle(QtGui.QApplication.translate("FindSpDialog", "Dialog", None, QtGui.QApplication.UnicodeUTF8))
        self.label.setText(QtGui.QApplication.translate("FindSpDialog", "Search", None, QtGui.QApplication.UnicodeUTF8))
        self.label_2.setText(QtGui.QApplication.translate("FindSpDialog", "Service", None, QtGui.QApplication.UnicodeUTF8))
        self.search.setText(QtGui.QApplication.translate("FindSpDialog", "Search", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setText(QtGui.QApplication.translate("FindSpDialog", "Add to dataset:", None, QtGui.QApplication.UnicodeUTF8))
        self.ds_add.setText(QtGui.QApplication.translate("FindSpDialog", "Add", None, QtGui.QApplication.UnicodeUTF8))

from PyQt4 import QtWebKit



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'taxonome/gui/ui/mainwindow.ui'
#
# Created: Tue May 29 11:46:18 2012
#      by: PyQt4 UI code generator 4.8.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName(_fromUtf8("MainWindow"))
        MainWindow.resize(709, 426)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(_fromUtf8(":/resources/leaf.svg")), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        MainWindow.setWindowIcon(icon)
        self.centralwidget = QtGui.QWidget(MainWindow)
        self.centralwidget.setObjectName(_fromUtf8("centralwidget"))
        self.verticalLayout_2 = QtGui.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.splitter_2 = QtGui.QSplitter(self.centralwidget)
        self.splitter_2.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_2.setObjectName(_fromUtf8("splitter_2"))
        self.verticalLayoutWidget = QtGui.QWidget(self.splitter_2)
        self.verticalLayoutWidget.setObjectName(_fromUtf8("verticalLayoutWidget"))
        self.verticalLayout_3 = QtGui.QVBoxLayout(self.verticalLayoutWidget)
        self.verticalLayout_3.setMargin(0)
        self.verticalLayout_3.setObjectName(_fromUtf8("verticalLayout_3"))
        self.label = QtGui.QLabel(self.verticalLayoutWidget)
        self.label.setObjectName(_fromUtf8("label"))
        self.verticalLayout_3.addWidget(self.label)
        self.datasets = QtGui.QListView(self.verticalLayoutWidget)
        self.datasets.setMinimumSize(QtCore.QSize(250, 0))
        self.datasets.setStyleSheet(_fromUtf8("QListView {font-size: 14pt}"))
        self.datasets.setObjectName(_fromUtf8("datasets"))
        self.verticalLayout_3.addWidget(self.datasets)
        self.layoutWidget = QtGui.QWidget(self.splitter_2)
        self.layoutWidget.setObjectName(_fromUtf8("layoutWidget"))
        self.verticalLayout = QtGui.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setMargin(0)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.search_label = QtGui.QLabel(self.layoutWidget)
        self.search_label.setObjectName(_fromUtf8("search_label"))
        self.verticalLayout.addWidget(self.search_label)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.searchbox = QtGui.QLineEdit(self.layoutWidget)
        self.searchbox.setObjectName(_fromUtf8("searchbox"))
        self.horizontalLayout_2.addWidget(self.searchbox)
        self.search_button = QtGui.QPushButton(self.layoutWidget)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.search_button.sizePolicy().hasHeightForWidth())
        self.search_button.setSizePolicy(sizePolicy)
        self.search_button.setObjectName(_fromUtf8("search_button"))
        self.horizontalLayout_2.addWidget(self.search_button)
        self.clearsearch_button = QtGui.QPushButton(self.layoutWidget)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Minimum, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.clearsearch_button.sizePolicy().hasHeightForWidth())
        self.clearsearch_button.setSizePolicy(sizePolicy)
        self.clearsearch_button.setMinimumSize(QtCore.QSize(0, 0))
        self.clearsearch_button.setObjectName(_fromUtf8("clearsearch_button"))
        self.horizontalLayout_2.addWidget(self.clearsearch_button)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.taxa_list_label = QtGui.QLabel(self.layoutWidget)
        self.taxa_list_label.setObjectName(_fromUtf8("taxa_list_label"))
        self.verticalLayout.addWidget(self.taxa_list_label)
        self.splitter = QtGui.QSplitter(self.layoutWidget)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName(_fromUtf8("splitter"))
        self.taxa_list = QtGui.QTableView(self.splitter)
        self.taxa_list.setMinimumSize(QtCore.QSize(0, 100))
        self.taxa_list.setObjectName(_fromUtf8("taxa_list"))
        self.tax_preview = QtWebKit.QWebView(self.splitter)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tax_preview.sizePolicy().hasHeightForWidth())
        self.tax_preview.setSizePolicy(sizePolicy)
        self.tax_preview.setMinimumSize(QtCore.QSize(300, 150))
        self.tax_preview.setUrl(QtCore.QUrl(_fromUtf8("about:blank")))
        self.tax_preview.setObjectName(_fromUtf8("tax_preview"))
        self.verticalLayout.addWidget(self.splitter)
        self.verticalLayout_2.addWidget(self.splitter_2)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtGui.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 709, 25))
        self.menubar.setObjectName(_fromUtf8("menubar"))
        self.menuFile = QtGui.QMenu(self.menubar)
        self.menuFile.setObjectName(_fromUtf8("menuFile"))
        self.menuTaxa = QtGui.QMenu(self.menubar)
        self.menuTaxa.setObjectName(_fromUtf8("menuTaxa"))
        self.menuWeb_services = QtGui.QMenu(self.menubar)
        self.menuWeb_services.setObjectName(_fromUtf8("menuWeb_services"))
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtGui.QStatusBar(MainWindow)
        self.statusbar.setObjectName(_fromUtf8("statusbar"))
        MainWindow.setStatusBar(self.statusbar)
        self.actionImport_CSV = QtGui.QAction(MainWindow)
        self.actionImport_CSV.setObjectName(_fromUtf8("actionImport_CSV"))
        self.actionLoad_JSONlines = QtGui.QAction(MainWindow)
        self.actionLoad_JSONlines.setObjectName(_fromUtf8("actionLoad_JSONlines"))
        self.actionSave_JSONlines = QtGui.QAction(MainWindow)
        self.actionSave_JSONlines.setObjectName(_fromUtf8("actionSave_JSONlines"))
        self.actionLoad_CSV_taxa = QtGui.QAction(MainWindow)
        self.actionLoad_CSV_taxa.setObjectName(_fromUtf8("actionLoad_CSV_taxa"))
        self.actionSave_CSV_taxa = QtGui.QAction(MainWindow)
        self.actionSave_CSV_taxa.setObjectName(_fromUtf8("actionSave_CSV_taxa"))
        self.actionMatch_names = QtGui.QAction(MainWindow)
        self.actionMatch_names.setObjectName(_fromUtf8("actionMatch_names"))
        self.actionCombine_datasets = QtGui.QAction(MainWindow)
        self.actionCombine_datasets.setObjectName(_fromUtf8("actionCombine_datasets"))
        self.actionLook_up_name = QtGui.QAction(MainWindow)
        self.actionLook_up_name.setObjectName(_fromUtf8("actionLook_up_name"))
        self.actionFetch_taxa = QtGui.QAction(MainWindow)
        self.actionFetch_taxa.setObjectName(_fromUtf8("actionFetch_taxa"))
        self.actionNew_ds = QtGui.QAction(MainWindow)
        self.actionNew_ds.setObjectName(_fromUtf8("actionNew_ds"))
        self.actionLoad_CSV_synonyms = QtGui.QAction(MainWindow)
        self.actionLoad_CSV_synonyms.setCheckable(False)
        self.actionLoad_CSV_synonyms.setChecked(False)
        self.actionLoad_CSV_synonyms.setObjectName(_fromUtf8("actionLoad_CSV_synonyms"))
        self.actionMap_big_csv = QtGui.QAction(MainWindow)
        self.actionMap_big_csv.setObjectName(_fromUtf8("actionMap_big_csv"))
        self.actionLoad_CSV_individuals = QtGui.QAction(MainWindow)
        self.actionLoad_CSV_individuals.setObjectName(_fromUtf8("actionLoad_CSV_individuals"))
        self.actionSave_CSV_synonyms = QtGui.QAction(MainWindow)
        self.actionSave_CSV_synonyms.setObjectName(_fromUtf8("actionSave_CSV_synonyms"))
        self.actionSave_CSV_individuals = QtGui.QAction(MainWindow)
        self.actionSave_CSV_individuals.setObjectName(_fromUtf8("actionSave_CSV_individuals"))
        self.actionPrevious_choices = QtGui.QAction(MainWindow)
        self.actionPrevious_choices.setObjectName(_fromUtf8("actionPrevious_choices"))
        self.menuFile.addAction(self.actionNew_ds)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionLoad_CSV_taxa)
        self.menuFile.addAction(self.actionLoad_CSV_synonyms)
        self.menuFile.addAction(self.actionLoad_CSV_individuals)
        self.menuFile.addAction(self.actionLoad_JSONlines)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.actionSave_CSV_taxa)
        self.menuFile.addAction(self.actionSave_CSV_synonyms)
        self.menuFile.addAction(self.actionSave_CSV_individuals)
        self.menuFile.addAction(self.actionSave_JSONlines)
        self.menuTaxa.addAction(self.actionMatch_names)
        self.menuTaxa.addAction(self.actionCombine_datasets)
        self.menuTaxa.addSeparator()
        self.menuTaxa.addAction(self.actionPrevious_choices)
        self.menuWeb_services.addAction(self.actionLook_up_name)
        self.menuWeb_services.addAction(self.actionFetch_taxa)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuTaxa.menuAction())
        self.menubar.addAction(self.menuWeb_services.menuAction())
        self.label.setBuddy(self.datasets)
        self.search_label.setBuddy(self.searchbox)
        self.taxa_list_label.setBuddy(self.taxa_list)

        self.retranslateUi(MainWindow)
        QtCore.QObject.connect(self.searchbox, QtCore.SIGNAL(_fromUtf8("returnPressed()")), MainWindow.do_search)
        QtCore.QObject.connect(self.search_button, QtCore.SIGNAL(_fromUtf8("clicked()")), MainWindow.do_search)
        QtCore.QObject.connect(self.clearsearch_button, QtCore.SIGNAL(_fromUtf8("clicked()")), self.searchbox.clear)
        QtCore.QObject.connect(self.clearsearch_button, QtCore.SIGNAL(_fromUtf8("clicked()")), MainWindow.clear_search)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QtGui.QApplication.translate("MainWindow", "Taxonome", None, QtGui.QApplication.UnicodeUTF8))
        self.label.setText(QtGui.QApplication.translate("MainWindow", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Ubuntu\'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600;\">Datasets</span></p></body></html>", None, QtGui.QApplication.UnicodeUTF8))
        self.search_label.setText(QtGui.QApplication.translate("MainWindow", "&Search taxa in this dataset", None, QtGui.QApplication.UnicodeUTF8))
        self.searchbox.setToolTip(QtGui.QApplication.translate("MainWindow", "Enter a species name, or use the * wildcard, e.g. \"Arachis *\".", None, QtGui.QApplication.UnicodeUTF8))
        self.search_button.setText(QtGui.QApplication.translate("MainWindow", "Search", None, QtGui.QApplication.UnicodeUTF8))
        self.clearsearch_button.setText(QtGui.QApplication.translate("MainWindow", "Clear", None, QtGui.QApplication.UnicodeUTF8))
        self.taxa_list_label.setText(QtGui.QApplication.translate("MainWindow", "Example taxa:", None, QtGui.QApplication.UnicodeUTF8))
        self.menuFile.setTitle(QtGui.QApplication.translate("MainWindow", "&File", None, QtGui.QApplication.UnicodeUTF8))
        self.menuTaxa.setTitle(QtGui.QApplication.translate("MainWindow", "&Taxa", None, QtGui.QApplication.UnicodeUTF8))
        self.menuWeb_services.setTitle(QtGui.QApplication.translate("MainWindow", "&Web services", None, QtGui.QApplication.UnicodeUTF8))
        self.actionImport_CSV.setText(QtGui.QApplication.translate("MainWindow", "Import CSV...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoad_JSONlines.setText(QtGui.QApplication.translate("MainWindow", "Load from JSONlines...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSave_JSONlines.setText(QtGui.QApplication.translate("MainWindow", "Save as JSONlines...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoad_CSV_taxa.setText(QtGui.QApplication.translate("MainWindow", "Load taxa from CSV...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoad_CSV_taxa.setToolTip(QtGui.QApplication.translate("MainWindow", "Taxa records: each species has one row.", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSave_CSV_taxa.setText(QtGui.QApplication.translate("MainWindow", "Save taxa to CSV...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionMatch_names.setText(QtGui.QApplication.translate("MainWindow", "Match taxa by name...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionCombine_datasets.setText(QtGui.QApplication.translate("MainWindow", "Combine datasets...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionCombine_datasets.setToolTip(QtGui.QApplication.translate("MainWindow", "Combine datasets that already use identical names.", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLook_up_name.setText(QtGui.QApplication.translate("MainWindow", "Look up name", None, QtGui.QApplication.UnicodeUTF8))
        self.actionFetch_taxa.setText(QtGui.QApplication.translate("MainWindow", "Fetch collection of taxa", None, QtGui.QApplication.UnicodeUTF8))
        self.actionNew_ds.setText(QtGui.QApplication.translate("MainWindow", "New empty dataset...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoad_CSV_synonyms.setText(QtGui.QApplication.translate("MainWindow", "Load synonymy from CSV...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoad_CSV_synonyms.setToolTip(QtGui.QApplication.translate("MainWindow", "Synonymy: Each name has one row, mapping it to an accepted name.", None, QtGui.QApplication.UnicodeUTF8))
        self.actionMap_big_csv.setText(QtGui.QApplication.translate("MainWindow", "Map names from big CSV dataset...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoad_CSV_individuals.setText(QtGui.QApplication.translate("MainWindow", "Load individual records from CSV...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionLoad_CSV_individuals.setToolTip(QtGui.QApplication.translate("MainWindow", "Individual records: a taxon may appear on more than one row - e.g. one row per specimen.", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSave_CSV_synonyms.setText(QtGui.QApplication.translate("MainWindow", "Save synonymy to CSV...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionSave_CSV_individuals.setText(QtGui.QApplication.translate("MainWindow", "Save individual records to CSV...", None, QtGui.QApplication.UnicodeUTF8))
        self.actionPrevious_choices.setText(QtGui.QApplication.translate("MainWindow", "Previous choices...", None, QtGui.QApplication.UnicodeUTF8))

from PyQt4 import QtWebKit
from . import icons_rc



from queue import Queue
from .qt import QtGui, QtCore

from taxonome.taxa.name_selector import NameSelector, TryAnotherName
from taxonome import Name

from .ui.nameselector import Ui_NameSelector
from .objects import NamePairItem

class GuiNameSelector(NameSelector, QtCore.QObject):
    choice_required = QtCore.Signal(object, object)
    choice_results = Queue()
    
    def present_dialog(self, name_options, name):
        # This must run in the main GUI thread.
        dialog = QtGui.QDialog()
        selectdialog = Ui_NameSelector()
        selectdialog.setupUi(dialog)
        
        if isinstance(name, Name):
            selectdialog.subject_line.setText("Matches for <i>{n.plain}</i> {n.authority}:"\
                                                .format(n=name))
        elif name:
            selectdialog.subject_line.setText("Matches for {}:".format(name))
        else:
            selectdialog.subject_line.setText("Matches:")
        
        choices = QtGui.QStandardItemModel()
        for namepair in name_options:
            choices.appendRow(NamePairItem(namepair))
        selectdialog.choice_list.setModel(choices)
        
        exitstatus = dialog.exec_()
        seln = selectdialog.choice_list.selectedIndexes()
        if seln:
            choice = choices.itemFromIndex(seln[0])
        else:
            choice = None
        newname = selectdialog.new_name.text()
        
        self.choice_results.put((exitstatus, choice, newname))
    
    def user_select(self, name_options, name, allow_retype, tracker):
        """Allow the user to choose between the options interactively."""
        def cache_choice(choice):
            self.store_user_choice(name_options, name, choice)
        
        self.choice_required.emit(name_options, name)
        exitstatus, choice, newname = self.choice_results.get()
        
        if not (exitstatus and (choice or newname)):
            ke = KeyError(name)
            cache_choice(ke)
            raise ke
        
        if newname:
            tracker.name_transform(name, newname, "user input")
            retry = TryAnotherName(newname)
            cache_choice(retry)
            raise retry
        
        cache_choice(choice.namepair)
        self.choice_results.task_done()
        return choice.namepair



import configparser
import os.path

default_config = {'main': {\
    'user-agent': 'Taxonome (in testing)',
    'name-fuzzy-threshold': '0.7',
    'author-fuzzy-threshold': '0.6',
    'load-tdwg-regions': 'True',
    },
'cache': {\
    'location': '~/.taxonome',
    'user-choices': '%(location)s/user_choices',
    'expiry-days': '30',
    'clean-on-load': 'True',
    },
'api-keys': {\
    # Programmatic users should get their own API keys.
    'tropicos': '',
    },
}

def load_config(filename="~/.taxonome/taxonome.cfg"):
    config = configparser.ConfigParser()
    config.read_dict(default_config)
    config.read([os.path.expanduser(filename)])
    return config

def save_config(tosave=None, filename="~/.taxonome/taxonome.cfg"):
    if tosave is None:
        tosave = config
    with open(os.path.expanduser(filename), 'w') as f:
        config.write(f)

config = load_config()
    



"""GUI Tools for doing things (e.g. loading big files) in a background thread.

Instead of blocking the GUI for long running operations, we do them in a
separate thread, and show a progress bar for them.
"""
from .qt import QtCore, QtGui
from functools import partial
import os

from taxonome.taxa import file_jsonlines, file_csv

class ProgressTracker(QtCore.QObject):
    """An object to signal the progress of a long running operation.
    
    Designed to be connected to a QProgressBar.
    """
    done_steps = QtCore.Signal(int)
    new_max = QtCore.Signal(int)
    def __call__(self, done):
        "Signal how many steps are complete."
        self.done_steps.emit(done)
    
    def max(self, n):
        self.new_max.emit(n)

class Worker(QtCore.QThread):
    """A simple worker thread. Runs the passed function.
    """
    error_raised = QtCore.Signal(object)
    def __init__(self, func, parent=None):
        super().__init__(parent)
        self.func = func
    
    def __del__(self):
        self.wait()
    
    def run(self):
        try:
            self.func()
        except Exception as e:
            self.error_raised.emit(e)
    
    def withprogress(self, app, steps=0):
        """Add a QProgressBar to the statusbar of app to track the progress
        of this thread. The progress bar will count from 0 to steps.
        """
        progressbar = QtGui.QProgressBar(app)
        progressbar.setRange(0, steps)
        progressbar.setMaximumSize(150, 20)
        app.ui.statusbar.addPermanentWidget(progressbar)
        
        if steps:
            progresstracker = ProgressTracker()
            progresstracker.done_steps.connect(progressbar.setValue)
            progresstracker.new_max.connect(progressbar.setMaximum)
            self.func = partial(self.func, progress=progresstracker)
        else:
            progresstracker = None
        
        def _stopped():
            app.ui.statusbar.removeWidget(progressbar)
            app.ui.statusbar.clearMessage()
            app.workerthreads.discard(self)
            
        self.finished.connect(_stopped)
        self.terminated.connect(_stopped)
        app.workerthreads.add(self)

class Loader(Worker):
    """A worker thread to load a dataset and send it via an 'output' signal."""
    output = QtCore.Signal(str, object)
    
    def __init__(self, func, ds_name, parent=None):
        super().__init__(func, parent)
        self.ds_name = ds_name
    
    def run(self):
        try:
            out = self.func()
        except Exception as e:
            self.error_raised.emit(e)
        else:
            self.output.emit(self.ds_name, out)

def makeloader(app, loadfunc, ds_name, steps=0):
    """Load a dataset into app using a separate thread, showing a progress bar."""
    thread = Loader(loadfunc, ds_name)
    thread.withprogress(app, steps)
    thread.output.connect(app.add_dataset)
    thread.error_raised.connect(app.show_error)
    thread.start()
    
def load_csv(app, ds_name, filename, encoding, namefield, authfield):
    steps = os.stat(filename).st_size
    fileobj = open(filename, encoding=encoding, errors='replace', newline='')
    loadfunc = partial(file_csv.load_taxa, fileobj, namefield=namefield,
                                                           authfield=authfield)
    app.ui.statusbar.showMessage("Loading taxa from {} ...".format(filename))
    return makeloader(app, loadfunc, ds_name, steps)

def load_csv_synonyms(app, ds_name, filename, encoding, **kwargs):
    steps = os.stat(filename).st_size
    fileobj = open(filename, encoding=encoding, errors='replace', newline='')
    loadfunc = partial(file_csv.load_synonyms, fileobj, **kwargs)
    app.ui.statusbar.showMessage("Loading synonymy from {} ...".format(filename))
    return makeloader(app, loadfunc, ds_name, steps)

def load_csv_individuals(app, ds_name, filename, encoding, **kwargs):
    steps = os.stat(filename).st_size
    fileobj = open(filename, encoding=encoding, errors='replace', newline='')
    loadfunc = partial(file_csv.load_individuals, fileobj, **kwargs)
    app.ui.statusbar.showMessage("Loading individual records from {} ...".format(filename))
    return makeloader(app, loadfunc, ds_name, steps)

def load_jsonlines(app, ds_name, filename):
    steps = os.stat(filename).st_size
    fileobj = open(filename, encoding='utf-8', errors='replace')
    loadfunc = partial(file_jsonlines.load_taxa, fileobj)
    app.ui.statusbar.showMessage("Loading taxa from {} ...".format(filename))
    return makeloader(app, loadfunc, ds_name, steps)



import unittest
import io
import warnings

from taxonome.taxa import file_csv
from taxonome import Name

csvdata = '''"Name","Author","Lives in"
"Solanum lycopersicum","L.","Peru"
"Lablab purpureus","(L.) Sweet.","Africa"'''

csvdata2 = '''"Name","Lives in"
"Solanum lycopersicum L.","Peru"
"Lablab purpureus (L.) Sweet.","Africa"'''

csvdata_noauth = '''"Name","Lives in"
"Solanum lycopersicum","Peru"
"Lablab purpureus","Africa"'''

csv_sp_data1 = io.StringIO('''"Name","Author"
"Solanum sp.",""
"Solanum spp.",""
"Solanum tuberosum","L."''')

csv_sp_data2 = io.StringIO('''"Name","Foo"
"Solanum sp.",""
"Solanum spp.",""
"Solanum tuberosum L.","Bar"''')

csv_sp_data3 = io.StringIO('''"Name","Foo"
"Solanum sp.",""
"Solanum spp.",""
"Solanum tuberosum","Bar"''')

class CsvTest(unittest.TestCase):
    def setUp(self):
        self.csvfile = io.StringIO(csvdata)
        self.ts = file_csv.load_taxa(self.csvfile)
        self.csvfile2 = io.StringIO(csvdata2)
        self.csvfile_noauth = io.StringIO(csvdata_noauth)
        
    def testRead(self):
        self.assertEqual(len(self.ts), 2)
        assert "Solanum lycopersicum" in self.ts
        assert Name("Lablab purpureus","Sweet") in self.ts
        self.assertEqual(self.ts["Lablab purpureus"].info["Lives in"], "Africa")
    
    def test_read_combined_name_auth(self):
        ts = file_csv.load_taxa(self.csvfile2, authfield=True)
        self.assertEqual(len(ts), 2)
        self.assertIn("Solanum lycopersicum", ts)
        self.assertIn(Name("Lablab purpureus","Sweet"), ts)
        self.assertEqual(ts["Solanum lycopersicum"].info["Lives in"], "Peru")
    
    def test_read_noauth(self):
        ts = file_csv.load_taxa(self.csvfile_noauth, authfield=None)
        self.assertEqual(len(ts), 2)
        self.assertIn("Solanum lycopersicum", ts)
        self.assertEqual(ts["Solanum lycopersicum"].info["Lives in"], "Peru")
    
    def testWrite(self):
        new_csvfile = io.StringIO()
        file_csv.save_taxa(new_csvfile, self.ts, info_fields=True)
        new_csvfile.seek(0)
        newts = file_csv.load_taxa(new_csvfile)
        self.assertEqual(len(newts), 2)
        self.assertEqual(newts["Lablab purpureus"].info["Lives in"], "Africa")
    
    def test_read_sp(self):
        "Check that 'sp.' and 'spp.' species are excluded when reading CSV."
        def check_ts(ts):
            self.assertEqual(len(ts), 1)
            self.assertIn("Solanum tuberosum", ts)
        
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            check_ts(file_csv.load_taxa(csv_sp_data1))
            check_ts(file_csv.load_taxa(csv_sp_data2, authfield=True))
            check_ts(file_csv.load_taxa(csv_sp_data3, authfield=None))

csv_syn_data = '''"Name","Author","Synonym name","Synonym author","Lives in"
"Solanum lycopersicum","L.","Lycopersicon lycopersicum","Foo.","Ignored"
"Solanum lycopersicum","L.","Solanum lycopersicum","L.","Peru"
"Lablab purpureus","(L.) Sweet.","Lablab purpureus","(L.) Sweet.","Africa"
"Lablab purpureus","(L.) Sweet.","Dolichos lablab","L.","Africa"
"Lablab purpureus","(L.) Sweet.","Vigna aristata","Piper","Ignored"
'''

class CSVSynonymyTests(unittest.TestCase):
    def setUp(self):
        self.csvfile = io.StringIO(csv_syn_data)
        self.ts = file_csv.load_synonyms(self.csvfile, synnamefield="Synonym name",
                                    synauthfield="Synonym author", accnamefield="Name",
                                    accauthfield="Author", get_info=True)
    
    def test_read(self):
        self.assertEqual(len(self.ts), 2)
        self.assertEqual(self.ts["Lablab purpureus"].othernames,
                {Name("Dolichos lablab","L."), Name("Vigna aristata", "Piper")})
        self.assertEqual(self.ts["Lablab purpureus"].info["Lives in"], "Africa")
        self.assertEqual(self.ts["Solanum lycopersicum"].info["Lives in"], "Peru")
        print(*self.ts.names)
        self.assertIn("Dolichos lablab", self.ts)
        self.assertIn("Lycopersicon lycopersicum", self.ts)
    
    def test_write(self):
        csvfile2 = io.StringIO()
        file_csv.save_synonyms(csvfile2, self.ts)
        csvfile2.seek(0)
        ts2 = file_csv.load_synonyms(csvfile2)
        self.assertEqual(len(ts2), 2)
        self.assertEqual(ts2["Lablab purpureus"].othernames,
                {Name("Dolichos lablab","L."), Name("Vigna aristata", "Piper")})
        print(*ts2.names)
        self.assertIn("Dolichos lablab", ts2)
        self.assertIn("Lycopersicon lycopersicum", ts2)

csv_individual_data = '''"Nom","Authority","Accession number","Seed mass"
"Vigna radiata","(L.) R. Wilczek","PI 291365",5.92
"Vigna radiata","(L.) R. Wilczek","PI 378024",5.27
"Vigna unguiculata","(L.) Walp.","Grif 13969",9.46
"Vigna unguiculata","(L.) Walp.","PI 349852 ",8.98
'''

class CSVIndividualsTest(unittest.TestCase):
    def setUp(self):
        self.csvfile = io.StringIO(csv_individual_data)
        self.ts = file_csv.load_individuals(self.csvfile, namefield="Nom", authfield="Authority")
    
    def check_ts(self, ts):
        self.assertEqual(len(ts), 2)
        Vr = ts["Vigna radiata"]
        self.assertEqual([ind["Accession number"] for ind in Vr.individuals], ["PI 291365", "PI 378024"])
        Vu = ts["Vigna unguiculata"]
        self.assertEqual([ind["Seed mass"] for ind in Vu.individuals], ["9.46", "8.98"])
    
    def test_read(self):
        self.check_ts(self.ts)
    
    def test_html(self):
        "Check that taxon.html() mentions individuals."
        for tax in self.ts:
            self.assertIn("individual records", tax.html())
    
    def test_write(self):
        csvfile2 = io.StringIO()
        file_csv.save_individuals(csvfile2, self.ts)
        csvfile2.seek(0)
        ts2 = file_csv.load_individuals(csvfile2)
        self.check_ts(ts2)
        



import urllib.request

from taxonome.config import config

USER_AGENT = config['main']['user-agent']

def urlopen(url, data=None):
    req = urllib.request.Request(url, data, headers={'ua': USER_AGENT})
    return urllib.request.urlopen(req)



import shelve, os, errno
import atexit
import warnings

import taxonome
from taxonome.config import config
from taxonome.tracker import noop_tracker

user_choice_file = os.path.expanduser(config['cache']['user-choices'])
try:
    os.makedirs(os.path.dirname(user_choice_file), exist_ok=True)
except OSError as e:
    if e.errno != errno.EEXIST:
        raise

user_choices = shelve.open(user_choice_file)
atexit.register(user_choices.close)

def no_op(*args): pass

class TryAnotherName(Exception):
    """This is raised if the user, when prompted, opts to try another name."""
    def __init__(self, newname):
        self.newname = newname
    def __str__(self):
        return "User asked to try another name, but code can't handle that here."

class AmbiguousMatchesError(KeyError):
    pass

class NameSelector:
    """Choose one from a list of possible matching names. If necessary, this
    will ask the user for interactive input. See the __call__ docstring for
    more details.
    
    This is used by the .select() and .optimistic_select() methods
    of TaxonSet objects."""
    def __init__(self, previous_choices=user_choices):
        """Set up the object for picking between alternative matching names.
        
        Parameters
        ----------
        previous_choices : mapping or None
          A mapping which can be used to record and repeat previous selections
          made by the user. By default, this uses a shelf, a persistent
          key-value store on disk. If None, this will be disabled.
        """
        super().__init__()
        self.previous_choices = previous_choices
    
    def __call__(self, name_options, name, allow_retype=True, tracker=noop_tracker):
        """Master function to select one name from a number of possibilities.
        
        Parameters
        ----------
        name_options : list
          A list of 2-tuples, (synonym, accepted_name), as returned by
          :meth:`TaxonSet.resolve_name`.
        name : Name or str
          The name we're trying to match.
        allow_retype : bool
          If True, offers the user the option to manually enter a new name. A
          TryAnotherName exception is raised; its 'newname' parameter contains
          the new name entered.
        tracker :
          A tracker object (see :mod:`taxonome.tracker`) to record the name
          selection process.
        
        Returns
        -------
        A 2-tuple, (synonym, accepted_name). Raises KeyError if name_options is
        empty or the user rejects all options. Raises TryAnotherName if the
        user chooses to enter a new name to search for.
        """
        rawname = name.plain if isinstance(name, taxonome.Name) else name
        
        def track_steps(selname, acc_name, main_method):
            """Calls tracker.name_transform for:
            - Fuzzy name matching (if applicable)
            - The selected name match, with main_method as the method
            - Syonymy (if matched name is not the accepted name)
            """
            if rawname != selname.plain:
                tracker.name_transform(rawname, selname.plain, "name variation")
            tracker.name_transform(name, selname, main_method)
            if selname != acc_name:
                tracker.name_transform(selname, acc_name, "synonymy")
        
        if not(name_options):                   # No options
            raise KeyError("No matching names", name)
            
        if len(name_options) == 1:              # One option (in list)
            n, an = name_options[0]
            track_steps(n, an, "one match")
            return name_options[0]
        
        an1 = name_options[0][1]                # All options are equivalent
        if all(an == an1 for n, an in name_options[1:]):
            n = name_options[0][0]
            track_steps(n, an1, "matches point to same taxon")
            return name_options[0]
        
        if self.previous_choices is not None and name is not None:
            choice = self.get_previous_choice(name_options, name, tracker)
            if choice is not None:
                track_steps(choice[0], choice[1], "recalled user selection")
                return choice
        
        try:
            choice = self.user_select(name_options, name, allow_retype=allow_retype,
                                    tracker=tracker)
        except NotImplementedError:
            warnings.warn("Ambiguous names, user selection not enabled.")
            raise AmbiguousMatchesError(name)
        
        track_steps(choice[0], choice[1], "user selection")
        return choice
        
    def get_previous_choice(self, name_options, name, tracker):
        try:
            prevoptions, prevchoice = self.previous_choices[repr(name)]
        except KeyError:
            return
        
        if prevoptions == set(name_options):
            if isinstance(prevchoice, Exception):
                # User selected one of the 'special' options.
                if isinstance(prevchoice, TryAnotherName):
                    tracker.name_transform(name, prevchoice.newname, "recalled user input")
                raise prevchoice
            # User selected one of the regular options:
            return prevchoice
    
    def store_user_choice(self, name_options, name, choice):
        if self.previous_choices is not None:
            self.previous_choices[repr(name)] = (set(name_options), choice)
        if hasattr(self.previous_choices, 'sync'):
            self.previous_choices.sync()
    
    def user_select(self, name_options, name, allow_retype, tracker):
        """This should be overridden by subclasses to let the user pick names
        interactively.
        
        Parameters:
          name_options: list
            Pairs of (matching_name, accepted_name).
          name: str
            The name that we're trying to match.
          allow_retype: bool
            If True, offer the user the option to type another name to search for.
          tracker:
            A tracker object - see details in :mod:`taxonome.tracker`.
        
        Returns:
          The pair of (matching_name, accepted_name) the user selected. Raises
          KeyError if the user rejected all choices.
        """
        raise NotImplementedError

class TerminalNameSelector(NameSelector):
    """Subclass of NameSelector presenting a terminal interface to let the
    user pick between alternative names."""
    def user_select(self, name_options, name, allow_retype, tracker):
        """Allow the user to choose between the options interactively."""
        def cache_choice(choice):
            self.store_user_choice(name_options, name, choice)
        
        if name:                            # Choose from multiple options
            print("== Select a match for {0}: ==".format(name))
        else:
            print("==Select a match: ==")
        print("N", "(Reject all)")
        if allow_retype:
            print(0, "(Specify new name to search for)")
        
        for i, (pname, pname_acc) in enumerate(name_options, start=1):
            if pname_acc and pname != pname_acc:
                print (i, pname, "\t  ->", pname_acc)
            else:
                print(i, pname)
                
        def _isvalid(inp):
            return (inp.isnumeric() and int(inp) >= 0 \
                        and int(inp) <= len(name_options)) \
                    or inp.upper() == "N"
        
        inp = input("Enter a number: ")
        while not _isvalid(inp):
            inp = input("Try again! Number: ")
        
        if inp.upper() == "N":
            ke = KeyError("User rejected all options", name)
            cache_choice(ke)
            raise ke
        if int(inp) == 0:
            newname = input("New name? ")
            tracker.name_transform(name, newname, "user input")
            retry = TryAnotherName(newname)
            cache_choice(retry)
            raise retry
        
        n, accname = name_options[int(inp)-1]
        cache_choice((n, accname))
        return n, accname



import re

from taxonome.config import config
FUZZY_COMPARISON_THRESHOLD = float(config['main']['author-fuzzy-threshold'])
from taxonome.utils import qgram

try:    # unidecode removes accents from letters to simplify comparison
    from unidecode import unidecode
except ImportError: # If we don't have unidecode, do nothing.
    def unidecode(input): return input
    
altspellings = {"Tsevelev":"Tzevelev"}

# f. seems to stand for filius, i.e. son. We currently ignore it.
surname_end_twiddles  = {"f."}
auth_noncapitalised = {"de", "du", "von", "van"}

class Author(object):
    initials = ()
    def __init__(self, author_str):
        *bits, surname = author_str.replace(".",". ").split()
        surname = surname.strip()
        # Special case for Linnaeus
        if surname == 'Linnaeus' and not bits:
            self.surname = 'L.'
            return

        if surname in surname_end_twiddles and bits:        # 'Forst. f.'
            *bits, surname = bits

        if bits and (bits[-1] in auth_noncapitalised):      # 'de Rigeur'
            *bits, prefix = bits
            surname = prefix + " " + surname

        self.initials = tuple(i[0].upper() for i in bits)
        self.surname = altspellings.get(surname, surname)
        
    def fuzzy_cmp(self, other):
        """Attempts a fuzzy logic comparison with another author name. The value
        returned is closer to 1 if they are the same, and closer to zero if they
        are different."""
        pdiff = 0.99
        sn1 = unidecode(self.surname)
        sn2 = unidecode(other.surname)
        if sn1.strip(".") == sn2.strip("."): # Exact match (give or take punctuation)
            pdiff *= 0.05
        elif min(len(sn1.strip(".")), len(sn2.strip("."))) > 1:
            # Try matching an abbreviation against a full name.
            if "." in sn1 and sn2.startswith(sn1.partition(".")[0]):
                pdiff *= 0.1
            if "." in sn2 and sn1.startswith(sn2.partition(".")[0]):
                pdiff *= 0.1
        if pdiff==0.99 and sn1.startswith(sn2[0]):
            # Fuzzy string matching with Q-grams if they start with the same
            # letter.
            score = qgram.compare_strings(sn1.strip("."), sn2.strip("."))
            pdiff *= 1 - (0.9 * score)
        # Bonus points for matching initials
        otherinitials = {unidecode(i) for i in other.initials}
        for i in self.initials:
            if unidecode(i) in otherinitials:
                pdiff *= 0.8
                
        return 1- pdiff
        
    def matches(self, other):
        return self.fuzzy_cmp(other) > FUZZY_COMPARISON_THRESHOLD

    def __eq__(self, other):
        return (self.initials == other.initials) and (self.surname == other.surname)
        
    def __ne__(self, other):
        return not(self.__eq__(other))
        
    def __repr__(self):
        return repr(self.initials) + " " + repr(self.surname)
        
    def __str__(self):
        if self.initials:
            part = " ".join(i+"." for i in self.initials)
            return part + " " + self.surname
        return self.surname
        
    def __hash__(self):
        return hash(self.initials) ^ hash(self.surname) ^ -311416313
        
    def __bool__(self):
        return bool(self.surname)
    
    def to_dict(self):
        return dict(initials=self.initials, surname=self.surname)

def _authors(s):
    return tuple(Author(n.strip()) for n in s.split("&") if n.strip() != "")

def _fuzzy_compare_groups(g1, g2, permatch_weight=1):
    if not(g1) or not(g2):
        return 1
    pdiff = 1.0
    for a1 in g1:
        pdiff *= 1- (permatch_weight * max(a1.fuzzy_cmp(a2) for a2 in g2))
    return pdiff

def _extract(pattern, auth_str, get_group=0):
    match = pattern.search(auth_str)
    if match:
        return match.group(get_group).strip(), pattern.sub("", auth_str)
    return "", auth_str
    
non_auth_re = re.compile(r"\bnon (.+)$")
prefix_re = re.compile(r"^(sensu)? ?(auct\.?)?")
suffix_re = re.compile(r"(, )?(p\.p\.)$")
basauth_re = re.compile(r"^\((.*)\)")
exauth_re = re.compile(r"\bex\.? (.*)")
year_re = re.compile(r",? +(\d{4})")

class Authority(object):
    basauth = ex = non = ()
    year = basyear = ""
    prefix = suffix = ""
    def __init__(self, auth_str):
        auth_str = auth_str.replace(" and "," & ").replace(" et al."," & al.").\
                    replace(" E. al.", " & al.")
        
        self.prefix, auth_str = _extract(prefix_re, auth_str)
        self.suffix, auth_str = _extract(suffix_re, auth_str, get_group=2)
        
        # "Person non Someoneelse"
        non_auth, auth_str = _extract(non_auth_re, auth_str, get_group=1)
        if non_auth:
            self.non = _authors(non_auth)
        auth_str = auth_str.strip(", ").replace(","," &")
        
        # p.p. suffix can be before or after "non Someoneelse".
        if not self.suffix:
            self.suffix, auth_str = _extract(suffix_re, auth_str, get_group=2)
        
        # "(Someone) Newauthority"
        basauth_str, auth_str = _extract(basauth_re, auth_str, get_group=1)
        if basauth_str:
            self.basyear, basauth_str = _extract(year_re, basauth_str, get_group=1)
            self.basauth = _authors(basauth_str.partition(" ex ")[0])
        
        # "Person, 1753"
        self.year, auth_str = _extract(year_re, auth_str, get_group=1)
        
        # "Person ex. Someoneelse"
        ex_str, auth_str =  _extract(exauth_re, auth_str, get_group=1)
        if ex_str:
            self.ex = _authors(ex_str)
            
        if auth_str.strip():
            self.main = _authors(auth_str)
            self._short = self._make_shortform() # For hash()
        else:
            self.main = ()
            self._short = ()
    
    def has_author(self, author):
        if any(a.matches(author) for a in self.non):
            return False
        if any(a.matches(author) for a in self.main + self.ex + self.basauth):
            return True
        return False
    
    def __iter__(self):
        return iter(self.main + self.ex + self.basauth)
    
    def non_match(self, other):
        """Returns True if any 'non X' authors match identified authors in the
        other taxon (either way round)."""
        if any(other.has_author(author) for author in self.non):
            return True
        if any(self.has_author(author) for author in other.non):
            return True
        return False
    
    def fuzzy_cmp(self, other):
        """Compares self with another Authority object, returning a number
        between 0 and 1 indicating the likelihood they match."""
        # Exact match:
        if self.main == other.main and self.ex == other.ex and self.basauth == other.basauth:
            return 1
        
        if self.non_match(other):
            return 0.01
        
        pdiff = 0.99
        pdiff *= _fuzzy_compare_groups(self.main, other.main, 0.9)
        if pdiff > 0.5: # Not matching yet
            if other.ex:
                pdiff *= _fuzzy_compare_groups(self.main, other.ex, 0.85)
            if self.ex:
                pdiff *= _fuzzy_compare_groups(self.ex, other.main, 0.85)
            return 1- pdiff
            
        if not(self.basauth) and not(other.basauth):
            pdiff *= 0.8
        elif self.basauth and other.basauth:
            pdiff *= _fuzzy_compare_groups(self.basauth, other.basauth, 0.5)
                
        if not(self.ex) and not(other.ex):
            pdiff *= 0.8
        elif self.ex and other.ex:
            pdiff *= _fuzzy_compare_groups(self.ex, other.ex, 0.5)
                
        return 1- pdiff
        
    def __eq__(self, other):
        return self.fuzzy_cmp(other) > FUZZY_COMPARISON_THRESHOLD
                
    def __ne__(self, other):
        return not(self.__eq__(other))
        
    def _make_shortform(self):   # Simplify authority.
        return tuple(unidecode(n.surname[:2].strip(".").lower()) for n in self.main)
        
    def __hash__(self):
        return hash(self.__class__) ^ hash(self._short)
                
#    def __hash__(self):
#        h = 1723863662
#        for a in self.main:
#            h ^= hash(a) ^ 359251845
#        if self.basauth:
#            for a in self.basauth:
#                h ^= hash(a) ^ -423436257
#        if self.ex:
#            for a in self.ex:
#                h ^= hash(a) ^ -289918925
#        return h
        
    def __repr__(self):
        return repr({"basauth":self.basauth, "main":self.main, "ex":self.ex,
                        "non": self.non})
        
    def __str__(self):
        basauth = mainauth = exauth = nonauth = None
        if self.basauth:
            basauth = "("+ " & ".join(str(a) for a in self.basauth)
            if self.basyear:
                basauth += ", " + self.basyear
            basauth += ")"
        if self.ex:
            exauth = "ex " + " & ".join(str(a) for a in self.ex)
        if self.non:
            nonauth = "non " + " & ".join(str(a) for a in self.non)
        if self.main:
            mainauth = " & ".join(str(a) for a in self.main)
        if self.year and mainauth:
            mainauth += ","
        bits = [self.prefix, basauth, mainauth, exauth, nonauth, self.year, self.suffix]
        return " ".join(b for b in bits if b)
        
    def __bool__(self):
        return bool(self.main or self.ex or self.basauth or self.non or \
                        self.prefix or self.suffix)
        
        



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'taxonome/gui/ui/prevchoices.ui'
#
# Created: Tue May 29 11:46:18 2012
#      by: PyQt4 UI code generator 4.8.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class Ui_PrevChoicesDialog(object):
    def setupUi(self, PrevChoicesDialog):
        PrevChoicesDialog.setObjectName(_fromUtf8("PrevChoicesDialog"))
        PrevChoicesDialog.resize(568, 475)
        self.verticalLayout_3 = QtGui.QVBoxLayout(PrevChoicesDialog)
        self.verticalLayout_3.setObjectName(_fromUtf8("verticalLayout_3"))
        self.splitter = QtGui.QSplitter(PrevChoicesDialog)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName(_fromUtf8("splitter"))
        self.layoutWidget = QtGui.QWidget(self.splitter)
        self.layoutWidget.setObjectName(_fromUtf8("layoutWidget"))
        self.verticalLayout = QtGui.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setMargin(0)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.ambignames = QtGui.QListView(self.layoutWidget)
        self.ambignames.setObjectName(_fromUtf8("ambignames"))
        self.verticalLayout.addWidget(self.ambignames)
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.forget = QtGui.QPushButton(self.layoutWidget)
        self.forget.setObjectName(_fromUtf8("forget"))
        self.horizontalLayout.addWidget(self.forget)
        self.forget_all = QtGui.QPushButton(self.layoutWidget)
        self.forget_all.setObjectName(_fromUtf8("forget_all"))
        self.horizontalLayout.addWidget(self.forget_all)
        spacerItem = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.layoutWidget1 = QtGui.QWidget(self.splitter)
        self.layoutWidget1.setObjectName(_fromUtf8("layoutWidget1"))
        self.verticalLayout_2 = QtGui.QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_2.setMargin(0)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.label = QtGui.QLabel(self.layoutWidget1)
        self.label.setObjectName(_fromUtf8("label"))
        self.verticalLayout_2.addWidget(self.label)
        self.chosenmatch = QtGui.QLabel(self.layoutWidget1)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.chosenmatch.setFont(font)
        self.chosenmatch.setText(_fromUtf8(""))
        self.chosenmatch.setScaledContents(False)
        self.chosenmatch.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.chosenmatch.setObjectName(_fromUtf8("chosenmatch"))
        self.verticalLayout_2.addWidget(self.chosenmatch)
        self.label_2 = QtGui.QLabel(self.layoutWidget1)
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.verticalLayout_2.addWidget(self.label_2)
        self.altmatches = QtGui.QListView(self.layoutWidget1)
        self.altmatches.setObjectName(_fromUtf8("altmatches"))
        self.verticalLayout_2.addWidget(self.altmatches)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.update_choice = QtGui.QPushButton(self.layoutWidget1)
        self.update_choice.setObjectName(_fromUtf8("update_choice"))
        self.horizontalLayout_2.addWidget(self.update_choice)
        self.reject_all = QtGui.QPushButton(self.layoutWidget1)
        self.reject_all.setObjectName(_fromUtf8("reject_all"))
        self.horizontalLayout_2.addWidget(self.reject_all)
        spacerItem1 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.verticalLayout_3.addWidget(self.splitter)
        self.buttonBox = QtGui.QDialogButtonBox(PrevChoicesDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Close)
        self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
        self.verticalLayout_3.addWidget(self.buttonBox)
        self.label_2.setBuddy(self.altmatches)

        self.retranslateUi(PrevChoicesDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), PrevChoicesDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), PrevChoicesDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(PrevChoicesDialog)

    def retranslateUi(self, PrevChoicesDialog):
        PrevChoicesDialog.setWindowTitle(QtGui.QApplication.translate("PrevChoicesDialog", "Stored name choices", None, QtGui.QApplication.UnicodeUTF8))
        self.forget.setText(QtGui.QApplication.translate("PrevChoicesDialog", "Forget", None, QtGui.QApplication.UnicodeUTF8))
        self.forget_all.setText(QtGui.QApplication.translate("PrevChoicesDialog", "Forget all", None, QtGui.QApplication.UnicodeUTF8))
        self.label.setText(QtGui.QApplication.translate("PrevChoicesDialog", "Preferred match:", None, QtGui.QApplication.UnicodeUTF8))
        self.label_2.setText(QtGui.QApplication.translate("PrevChoicesDialog", "Alternatives:", None, QtGui.QApplication.UnicodeUTF8))
        self.update_choice.setText(QtGui.QApplication.translate("PrevChoicesDialog", "Update choice", None, QtGui.QApplication.UnicodeUTF8))
        self.reject_all.setText(QtGui.QApplication.translate("PrevChoicesDialog", "Reject all", None, QtGui.QApplication.UnicodeUTF8))




#!/usr/bin/python3
# -*- coding: utf-8 -*-
import shelve
import datetime
DEFAULT_EXPIRY = datetime.timedelta(days=7)

class ForgetfulShelf(shelve.DbfilenameShelf):
    def __init__(self, filename, flag='c', protocol=None, writeback=False,
                    expiry=DEFAULT_EXPIRY):
        super().__init__(filename, flag, protocol, writeback)
        self.expiry = expiry
        
    def __setitem__(self, key, value):
        value = (value, datetime.datetime.now())
        super().__setitem__(key, value)
    
    def __getitem__(self, key):
        value, timestamp = super().__getitem__(key)
        if (datetime.datetime.now() - timestamp) > self.expiry:
            del self[key]
            raise KeyError("Item expired in cache")
        return value
        
    def _get_with_timestamp(self, key):
        """Returns a 2-tuple of the value and the timestamp (as a
        datetime.datetime object).
        
        This will work even if the item has expired, but only if it hasn't
        been deleted. Attempting to access expired items with Cache[key] will
        delete them.
        """
        return super().__getitem__(key)
    
    def _set_with_timestamp(self, key, value, timestamp):
        """Override the automatic timestamping, using a datetime.datetime."""
        super().__setitem__(key, (value, timestamp))
    
    def _get_timestamp(self, key):
        """Just return the timestamp of a key in the database (whether or not
        it has expired).
        
        Will raise a KeyError if the key isn't there."""
        return self._get_with_timestamp(key)[1]
    
    def __contains__(self, key):
        if super().__contains__(key) and \
            (datetime.datetime.now() - self._get_timestamp(key)) < self.expiry:
                return True
        return False
    
    def append_to(self, key, value):
        """Add an item to a mutable list on the shelf. Simply retrieving
        and modifying a value will not write it back to the disk.
        
        If the key is not found on the shelf, it will be created, with
        the value in a new list.
        
        The timestamp will remain that of the oldest item."""
        if key in self:
            temp, timestamp = self._get_with_timestamp(key)
            temp.append(value)
                # Setting self[key] would use a new timestamp, so:
            self._set_with_timestamp(key, temp, timestamp)
        else: # Start a new list
            self[key] = [value]
        
    def clean_outdated(self):
        """Delete all items on the Shelf older than the expiry time. Returns
        the number of items deleted.
        
        This may be slow, as it scans every item in the database."""
        deleted = 0
        for key in self:
            print(key)
            if (datetime.datetime.now() - self._get_timestamp(key)) > self.expiry:
                del self[key]
                deleted += 1
        return deleted
                        


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'taxonome/gui/ui/csvimport_synonyms.ui'
#
# Created: Tue May 29 11:46:18 2012
#      by: PyQt4 UI code generator 4.8.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class Ui_CsvSynonymsImportDialog(object):
    def setupUi(self, CsvSynonymsImportDialog):
        CsvSynonymsImportDialog.setObjectName(_fromUtf8("CsvSynonymsImportDialog"))
        CsvSynonymsImportDialog.resize(551, 647)
        self.verticalLayout_3 = QtGui.QVBoxLayout(CsvSynonymsImportDialog)
        self.verticalLayout_3.setObjectName(_fromUtf8("verticalLayout_3"))
        self.horizontalLayout_3 = QtGui.QHBoxLayout()
        self.horizontalLayout_3.setObjectName(_fromUtf8("horizontalLayout_3"))
        self.label_3 = QtGui.QLabel(CsvSynonymsImportDialog)
        self.label_3.setTextFormat(QtCore.Qt.AutoText)
        self.label_3.setObjectName(_fromUtf8("label_3"))
        self.horizontalLayout_3.addWidget(self.label_3)
        self.ds_name = QtGui.QLineEdit(CsvSynonymsImportDialog)
        self.ds_name.setObjectName(_fromUtf8("ds_name"))
        self.horizontalLayout_3.addWidget(self.ds_name)
        self.verticalLayout_3.addLayout(self.horizontalLayout_3)
        self.accepted_names = QtGui.QGroupBox(CsvSynonymsImportDialog)
        self.accepted_names.setFlat(False)
        self.accepted_names.setCheckable(False)
        self.accepted_names.setObjectName(_fromUtf8("accepted_names"))
        self.verticalLayout = QtGui.QVBoxLayout(self.accepted_names)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.label = QtGui.QLabel(self.accepted_names)
        self.label.setTextFormat(QtCore.Qt.AutoText)
        self.label.setObjectName(_fromUtf8("label"))
        self.horizontalLayout.addWidget(self.label)
        self.accnamefield = QtGui.QComboBox(self.accepted_names)
        self.accnamefield.setMinimumSize(QtCore.QSize(150, 0))
        self.accnamefield.setObjectName(_fromUtf8("accnamefield"))
        self.horizontalLayout.addWidget(self.accnamefield)
        spacerItem = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.groupBox = QtGui.QGroupBox(self.accepted_names)
        self.groupBox.setObjectName(_fromUtf8("groupBox"))
        self.formLayout_2 = QtGui.QFormLayout(self.groupBox)
        self.formLayout_2.setObjectName(_fromUtf8("formLayout_2"))
        self.accnoauth = QtGui.QRadioButton(self.groupBox)
        self.accnoauth.setObjectName(_fromUtf8("accnoauth"))
        self.formLayout_2.setWidget(0, QtGui.QFormLayout.LabelRole, self.accnoauth)
        self.accauth_with_name = QtGui.QRadioButton(self.groupBox)
        self.accauth_with_name.setObjectName(_fromUtf8("accauth_with_name"))
        self.formLayout_2.setWidget(1, QtGui.QFormLayout.LabelRole, self.accauth_with_name)
        self.accauth_separate = QtGui.QRadioButton(self.groupBox)
        self.accauth_separate.setChecked(True)
        self.accauth_separate.setObjectName(_fromUtf8("accauth_separate"))
        self.formLayout_2.setWidget(3, QtGui.QFormLayout.LabelRole, self.accauth_separate)
        self.accauthfield = QtGui.QComboBox(self.groupBox)
        self.accauthfield.setEnabled(True)
        self.accauthfield.setObjectName(_fromUtf8("accauthfield"))
        self.formLayout_2.setWidget(3, QtGui.QFormLayout.FieldRole, self.accauthfield)
        self.verticalLayout.addWidget(self.groupBox)
        self.verticalLayout_3.addWidget(self.accepted_names)
        self.synonyms_grp = QtGui.QGroupBox(CsvSynonymsImportDialog)
        self.synonyms_grp.setObjectName(_fromUtf8("synonyms_grp"))
        self.verticalLayout_2 = QtGui.QVBoxLayout(self.synonyms_grp)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.label_4 = QtGui.QLabel(self.synonyms_grp)
        self.label_4.setTextFormat(QtCore.Qt.AutoText)
        self.label_4.setObjectName(_fromUtf8("label_4"))
        self.horizontalLayout_2.addWidget(self.label_4)
        self.synnamefield = QtGui.QComboBox(self.synonyms_grp)
        self.synnamefield.setMinimumSize(QtCore.QSize(150, 0))
        self.synnamefield.setObjectName(_fromUtf8("synnamefield"))
        self.horizontalLayout_2.addWidget(self.synnamefield)
        spacerItem1 = QtGui.QSpacerItem(40, 20, QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.groupBox_4 = QtGui.QGroupBox(self.synonyms_grp)
        self.groupBox_4.setObjectName(_fromUtf8("groupBox_4"))
        self.formLayout_3 = QtGui.QFormLayout(self.groupBox_4)
        self.formLayout_3.setObjectName(_fromUtf8("formLayout_3"))
        self.synnoauth = QtGui.QRadioButton(self.groupBox_4)
        self.synnoauth.setObjectName(_fromUtf8("synnoauth"))
        self.formLayout_3.setWidget(0, QtGui.QFormLayout.LabelRole, self.synnoauth)
        self.synauth_with_name = QtGui.QRadioButton(self.groupBox_4)
        self.synauth_with_name.setObjectName(_fromUtf8("synauth_with_name"))
        self.formLayout_3.setWidget(1, QtGui.QFormLayout.LabelRole, self.synauth_with_name)
        self.synauth_separate = QtGui.QRadioButton(self.groupBox_4)
        self.synauth_separate.setChecked(True)
        self.synauth_separate.setObjectName(_fromUtf8("synauth_separate"))
        self.formLayout_3.setWidget(3, QtGui.QFormLayout.LabelRole, self.synauth_separate)
        self.synauthfield = QtGui.QComboBox(self.groupBox_4)
        self.synauthfield.setEnabled(True)
        self.synauthfield.setObjectName(_fromUtf8("synauthfield"))
        self.formLayout_3.setWidget(3, QtGui.QFormLayout.FieldRole, self.synauthfield)
        self.verticalLayout_2.addWidget(self.groupBox_4)
        self.verticalLayout_3.addWidget(self.synonyms_grp)
        self.label_2 = QtGui.QLabel(CsvSynonymsImportDialog)
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.verticalLayout_3.addWidget(self.label_2)
        self.preview = QtGui.QTableWidget(CsvSynonymsImportDialog)
        self.preview.setObjectName(_fromUtf8("preview"))
        self.preview.setColumnCount(0)
        self.preview.setRowCount(0)
        self.verticalLayout_3.addWidget(self.preview)
        self.horizontalLayout_4 = QtGui.QHBoxLayout()
        self.horizontalLayout_4.setObjectName(_fromUtf8("horizontalLayout_4"))
        self.label_5 = QtGui.QLabel(CsvSynonymsImportDialog)
        self.label_5.setObjectName(_fromUtf8("label_5"))
        self.horizontalLayout_4.addWidget(self.label_5)
        self.csv_encoding = QtGui.QComboBox(CsvSynonymsImportDialog)
        self.csv_encoding.setObjectName(_fromUtf8("csv_encoding"))
        self.horizontalLayout_4.addWidget(self.csv_encoding)
        self.verticalLayout_3.addLayout(self.horizontalLayout_4)
        self.buttonBox = QtGui.QDialogButtonBox(CsvSynonymsImportDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
        self.verticalLayout_3.addWidget(self.buttonBox)
        self.label_3.setBuddy(self.ds_name)
        self.label.setBuddy(self.accnamefield)
        self.label_4.setBuddy(self.synnamefield)
        self.label_2.setBuddy(self.preview)

        self.retranslateUi(CsvSynonymsImportDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), CsvSynonymsImportDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), CsvSynonymsImportDialog.reject)
        QtCore.QObject.connect(self.accauth_separate, QtCore.SIGNAL(_fromUtf8("toggled(bool)")), self.accauthfield.setEnabled)
        QtCore.QMetaObject.connectSlotsByName(CsvSynonymsImportDialog)

    def retranslateUi(self, CsvSynonymsImportDialog):
        CsvSynonymsImportDialog.setWindowTitle(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Import CSV", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Ubuntu\'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600;\">Dataset name:</span></p></body></html>", None, QtGui.QApplication.UnicodeUTF8))
        self.accepted_names.setTitle(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Accepted names", None, QtGui.QApplication.UnicodeUTF8))
        self.label.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Name column:", None, QtGui.QApplication.UnicodeUTF8))
        self.groupBox.setTitle(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Authorities:", None, QtGui.QApplication.UnicodeUTF8))
        self.accnoauth.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "None", None, QtGui.QApplication.UnicodeUTF8))
        self.accauth_with_name.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "In column with species", None, QtGui.QApplication.UnicodeUTF8))
        self.accauth_separate.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "In separate column:", None, QtGui.QApplication.UnicodeUTF8))
        self.synonyms_grp.setTitle(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Synonyms", None, QtGui.QApplication.UnicodeUTF8))
        self.label_4.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Name column:", None, QtGui.QApplication.UnicodeUTF8))
        self.groupBox_4.setTitle(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Authorities:", None, QtGui.QApplication.UnicodeUTF8))
        self.synnoauth.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "None", None, QtGui.QApplication.UnicodeUTF8))
        self.synauth_with_name.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "In column with species", None, QtGui.QApplication.UnicodeUTF8))
        self.synauth_separate.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "In separate column:", None, QtGui.QApplication.UnicodeUTF8))
        self.label_2.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Ubuntu\'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600;\">Preview</span></p></body></html>", None, QtGui.QApplication.UnicodeUTF8))
        self.label_5.setText(QtGui.QApplication.translate("CsvSynonymsImportDialog", "Character encoding", None, QtGui.QApplication.UnicodeUTF8))




# -*- coding: utf-8 -*-
"""This module fetches information from the World Register of Marine Species
webservice.

http://www.marinespecies.org/

Technical note: tested using the Python 3 patches for suds available at
https://bitbucket.org/bernh/suds-python-3-patches/wiki/Home
"""
from suds.client import Client

from taxonome import Name, Taxon
from taxonome.taxa.collection import TaxaResource, TaxonSet

WSDL_URL = "http://www.marinespecies.org/aphia.php?p=soap&wsdl=1"

def _make_taxon(record):
    tax = Taxon(record.valid_name, record.valid_authority)
    tax.name.aphia_id = record.valid_AphiaID
    
    if record.status == "accepted":
        tax.lsid = record.lsid
        tax.url = record.url
    else:
        syn = Name(record.scientficname, record.authority)
        tax.othernames.add(syn)
    
    return tax

class WoRMSTaxaResource(TaxaResource):
    def __init__(self):
        super().__init__()
        self.service = Client(WSDL_URL)
        self.cache_by_id = {}
        
    def get_by_accepted_name(self, name):
        try:
            tax = self.cache_by_id[name.aphia_id]
        except (AttributeError, KeyError):
            records = self.service.service.getAphiaRecords(name.plain)
            tax = _make_taxon(records[0])
        
        for synrecord in self.service.service.getAphiaSynonymsByID(tax.name.aphia_id):
            syn = Name(synrecord.scientificname, synrecord.authority)
            tax.othernames.add(syn)
        
        self.cache_by_id[tax.name.aphia_id] = tax
        return tax
    
    def resolve_name(self, key):
        if isinstance(key, Name):
            return [(n, an) for n, an in self.resolve_name(key.plain) \
                                            if n.authority==key.authority]
        
        names = []
        for record in self.service.service.getAphiaRecords(key):
            n = Name(record.scientificname, record.authority)
            n.aphia_id = record.AphiaID
            an = Name(record.valid_name, record.valid_authority)
            an.aphia_id = record.AphiaID
            names.append((n, an))
        return names

_instance = None
def _get_instance():
    global _instance
    if not _instance:
        _instance = WoRMSTaxaResource()
    return _instance

def select(*args, **kwargs):
    return _get_instance().select(*args, **kwargs)

def resolve_name(*args, **kwargs):
    return _get_instance().resolve_name(*args, **kwargs)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'taxonome/gui/ui/combine.ui'
#
# Created: Tue May 29 11:46:18 2012
#      by: PyQt4 UI code generator 4.8.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class Ui_CombineDialog(object):
    def setupUi(self, CombineDialog):
        CombineDialog.setObjectName(_fromUtf8("CombineDialog"))
        CombineDialog.resize(587, 460)
        self.verticalLayout_2 = QtGui.QVBoxLayout(CombineDialog)
        self.verticalLayout_2.setObjectName(_fromUtf8("verticalLayout_2"))
        self.horizontalLayout = QtGui.QHBoxLayout()
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.label = QtGui.QLabel(CombineDialog)
        self.label.setObjectName(_fromUtf8("label"))
        self.horizontalLayout.addWidget(self.label)
        self.newname = QtGui.QLineEdit(CombineDialog)
        self.newname.setObjectName(_fromUtf8("newname"))
        self.horizontalLayout.addWidget(self.newname)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.groupBox = QtGui.QGroupBox(CombineDialog)
        self.groupBox.setObjectName(_fromUtf8("groupBox"))
        self.verticalLayout_6 = QtGui.QVBoxLayout(self.groupBox)
        self.verticalLayout_6.setObjectName(_fromUtf8("verticalLayout_6"))
        self.splitter = QtGui.QSplitter(self.groupBox)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName(_fromUtf8("splitter"))
        self.layoutWidget = QtGui.QWidget(self.splitter)
        self.layoutWidget.setObjectName(_fromUtf8("layoutWidget"))
        self.verticalLayout = QtGui.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setMargin(0)
        self.verticalLayout.setObjectName(_fromUtf8("verticalLayout"))
        self.horizontalLayout_3 = QtGui.QHBoxLayout()
        self.horizontalLayout_3.setObjectName(_fromUtf8("horizontalLayout_3"))
        self.label_2 = QtGui.QLabel(self.layoutWidget)
        self.label_2.setObjectName(_fromUtf8("label_2"))
        self.horizontalLayout_3.addWidget(self.label_2)
        self.target_ds_add = QtGui.QToolButton(self.layoutWidget)
        self.target_ds_add.setAutoRaise(True)
        self.target_ds_add.setObjectName(_fromUtf8("target_ds_add"))
        self.horizontalLayout_3.addWidget(self.target_ds_add)
        self.target_ds_remove = QtGui.QToolButton(self.layoutWidget)
        self.target_ds_remove.setAutoRaise(True)
        self.target_ds_remove.setObjectName(_fromUtf8("target_ds_remove"))
        self.horizontalLayout_3.addWidget(self.target_ds_remove)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.target_ds_list = QtGui.QListView(self.layoutWidget)
        self.target_ds_list.setObjectName(_fromUtf8("target_ds_list"))
        self.verticalLayout.addWidget(self.target_ds_list)
        self.horizontalLayout_4 = QtGui.QHBoxLayout()
        self.horizontalLayout_4.setObjectName(_fromUtf8("horizontalLayout_4"))
        self.label_3 = QtGui.QLabel(self.layoutWidget)
        self.label_3.setObjectName(_fromUtf8("label_3"))
        self.horizontalLayout_4.addWidget(self.label_3)
        self.bg_ds_add = QtGui.QToolButton(self.layoutWidget)
        self.bg_ds_add.setAutoRaise(True)
        self.bg_ds_add.setObjectName(_fromUtf8("bg_ds_add"))
        self.horizontalLayout_4.addWidget(self.bg_ds_add)
        self.bg_ds_remove = QtGui.QToolButton(self.layoutWidget)
        self.bg_ds_remove.setAutoRaise(True)
        self.bg_ds_remove.setArrowType(QtCore.Qt.NoArrow)
        self.bg_ds_remove.setObjectName(_fromUtf8("bg_ds_remove"))
        self.horizontalLayout_4.addWidget(self.bg_ds_remove)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.bg_ds_list = QtGui.QListView(self.layoutWidget)
        self.bg_ds_list.setObjectName(_fromUtf8("bg_ds_list"))
        self.verticalLayout.addWidget(self.bg_ds_list)
        self.layoutWidget1 = QtGui.QWidget(self.splitter)
        self.layoutWidget1.setObjectName(_fromUtf8("layoutWidget1"))
        self.verticalLayout_5 = QtGui.QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_5.setMargin(0)
        self.verticalLayout_5.setObjectName(_fromUtf8("verticalLayout_5"))
        self.label_4 = QtGui.QLabel(self.layoutWidget1)
        self.label_4.setObjectName(_fromUtf8("label_4"))
        self.verticalLayout_5.addWidget(self.label_4)
        self.available_ds_list = QtGui.QListView(self.layoutWidget1)
        self.available_ds_list.setObjectName(_fromUtf8("available_ds_list"))
        self.verticalLayout_5.addWidget(self.available_ds_list)
        self.verticalLayout_6.addWidget(self.splitter)
        self.verticalLayout_2.addWidget(self.groupBox)
        self.horizontalLayout_2 = QtGui.QHBoxLayout()
        self.horizontalLayout_2.setObjectName(_fromUtf8("horizontalLayout_2"))
        self.label_5 = QtGui.QLabel(CombineDialog)
        self.label_5.setObjectName(_fromUtf8("label_5"))
        self.horizontalLayout_2.addWidget(self.label_5)
        self.distrib_ds = QtGui.QComboBox(CombineDialog)
        sizePolicy = QtGui.QSizePolicy(QtGui.QSizePolicy.Expanding, QtGui.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.distrib_ds.sizePolicy().hasHeightForWidth())
        self.distrib_ds.setSizePolicy(sizePolicy)
        self.distrib_ds.setEditable(False)
        self.distrib_ds.setObjectName(_fromUtf8("distrib_ds"))
        self.horizontalLayout_2.addWidget(self.distrib_ds)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.nest_info = QtGui.QCheckBox(CombineDialog)
        self.nest_info.setObjectName(_fromUtf8("nest_info"))
        self.verticalLayout_2.addWidget(self.nest_info)
        self.buttonBox = QtGui.QDialogButtonBox(CombineDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
        self.verticalLayout_2.addWidget(self.buttonBox)
        self.label.setBuddy(self.newname)
        self.label_5.setBuddy(self.distrib_ds)

        self.retranslateUi(CombineDialog)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), CombineDialog.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), CombineDialog.reject)
        QtCore.QObject.connect(self.bg_ds_add, QtCore.SIGNAL(_fromUtf8("clicked()")), CombineDialog.add_bg_ds)
        QtCore.QObject.connect(self.bg_ds_remove, QtCore.SIGNAL(_fromUtf8("clicked()")), CombineDialog.remove_bg_ds)
        QtCore.QObject.connect(self.target_ds_add, QtCore.SIGNAL(_fromUtf8("clicked()")), CombineDialog.add_target_ds)
        QtCore.QObject.connect(self.target_ds_remove, QtCore.SIGNAL(_fromUtf8("clicked()")), CombineDialog.remove_target_ds)
        QtCore.QMetaObject.connectSlotsByName(CombineDialog)

    def retranslateUi(self, CombineDialog):
        CombineDialog.setWindowTitle(QtGui.QApplication.translate("CombineDialog", "Combine datasets", None, QtGui.QApplication.UnicodeUTF8))
        self.label.setText(QtGui.QApplication.translate("CombineDialog", "&New name:", None, QtGui.QApplication.UnicodeUTF8))
        self.groupBox.setTitle(QtGui.QApplication.translate("CombineDialog", "Pick datasets to combine", None, QtGui.QApplication.UnicodeUTF8))
        self.label_2.setToolTip(QtGui.QApplication.translate("CombineDialog", "All taxa from each target dataset will be included.", None, QtGui.QApplication.UnicodeUTF8))
        self.label_2.setText(QtGui.QApplication.translate("CombineDialog", "Target datasets", None, QtGui.QApplication.UnicodeUTF8))
        self.target_ds_add.setText(QtGui.QApplication.translate("CombineDialog", " + ", None, QtGui.QApplication.UnicodeUTF8))
        self.target_ds_remove.setText(QtGui.QApplication.translate("CombineDialog", " - ", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setToolTip(QtGui.QApplication.translate("CombineDialog", "Information from background datasets will only be used for taxa in target datasets.", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setText(QtGui.QApplication.translate("CombineDialog", "Background datasets", None, QtGui.QApplication.UnicodeUTF8))
        self.bg_ds_add.setText(QtGui.QApplication.translate("CombineDialog", " + ", None, QtGui.QApplication.UnicodeUTF8))
        self.bg_ds_remove.setText(QtGui.QApplication.translate("CombineDialog", " - ", None, QtGui.QApplication.UnicodeUTF8))
        self.label_4.setText(QtGui.QApplication.translate("CombineDialog", "Available datasets", None, QtGui.QApplication.UnicodeUTF8))
        self.label_5.setText(QtGui.QApplication.translate("CombineDialog", "Dataset for &distribution:", None, QtGui.QApplication.UnicodeUTF8))
        self.nest_info.setText(QtGui.QApplication.translate("CombineDialog", "Nest information fields by source dataset (leave off if you want to save as CSV)", None, QtGui.QApplication.UnicodeUTF8))




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'taxonome/gui/ui/nameselector.ui'
#
# Created: Tue May 29 11:46:18 2012
#      by: PyQt4 UI code generator 4.8.3
#
# WARNING! All changes made in this file will be lost!

from PyQt4 import QtCore, QtGui

try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

class Ui_NameSelector(object):
    def setupUi(self, NameSelector):
        NameSelector.setObjectName(_fromUtf8("NameSelector"))
        NameSelector.resize(431, 400)
        self.buttonBox = QtGui.QDialogButtonBox(NameSelector)
        self.buttonBox.setGeometry(QtCore.QRect(80, 360, 341, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtGui.QDialogButtonBox.Cancel|QtGui.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName(_fromUtf8("buttonBox"))
        self.subject_line = QtGui.QLabel(NameSelector)
        self.subject_line.setGeometry(QtCore.QRect(20, 20, 361, 17))
        self.subject_line.setObjectName(_fromUtf8("subject_line"))
        self.choice_list = QtGui.QListView(NameSelector)
        self.choice_list.setGeometry(QtCore.QRect(20, 50, 391, 231))
        self.choice_list.setEditTriggers(QtGui.QAbstractItemView.NoEditTriggers)
        self.choice_list.setObjectName(_fromUtf8("choice_list"))
        self.label_3 = QtGui.QLabel(NameSelector)
        self.label_3.setGeometry(QtCore.QRect(20, 300, 331, 17))
        self.label_3.setObjectName(_fromUtf8("label_3"))
        self.new_name = QtGui.QLineEdit(NameSelector)
        self.new_name.setGeometry(QtCore.QRect(20, 320, 391, 27))
        self.new_name.setObjectName(_fromUtf8("new_name"))
        self.label_3.setBuddy(self.new_name)

        self.retranslateUi(NameSelector)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("accepted()")), NameSelector.accept)
        QtCore.QObject.connect(self.buttonBox, QtCore.SIGNAL(_fromUtf8("rejected()")), NameSelector.reject)
        QtCore.QMetaObject.connectSlotsByName(NameSelector)

    def retranslateUi(self, NameSelector):
        NameSelector.setWindowTitle(QtGui.QApplication.translate("NameSelector", "Select matching name", None, QtGui.QApplication.UnicodeUTF8))
        self.subject_line.setText(QtGui.QApplication.translate("NameSelector", "Matches for [placeholder]:", None, QtGui.QApplication.UnicodeUTF8))
        self.label_3.setText(QtGui.QApplication.translate("NameSelector", "Or type another name to find as a &replacement:", None, QtGui.QApplication.UnicodeUTF8))




base_url = "http://tnrs.iplantc.org/tnrsm-svc/matchNames?"

import io
from itertools import chain
import json
from urllib.parse import urlencode

from taxonome import Name, Taxon
from taxonome.taxa.collection import build_matched_taxonset, combine_dicts
from taxonome.tracker import noop_tracker, prepare_tracker
from .utils import urlopen

def call_api(names, retrieve="best"):
    names = ",".join(names)
    query = urlencode(dict(names=names, retrieve=retrieve))
    return json.load(io.TextIOWrapper(urlopen(base_url + query), 'utf-8'))

def _make_name(result_dict):
    if not result_dict['acceptedName']:
        return None
    return Name(result_dict['acceptedName'], result_dict['authorAttributed'])

def _make_taxon(result_dict):
    n = _make_name(result_dict)
    if n is None:
        return None
    t = Taxon(n)
    t.info['family'] = result_dict['family']
    return t

def chunklist(l, chunksize=100):
    start=0
    while start < len(l):
        chunkend = min(start+chunksize, len(l))
        # Work around an apparent bug in TNRS: if a name has '&' in it (for
        # authorities), TNRS stops processing at that point.
        for i in range(start, chunkend):
            if '&' in l[i]:
                yield l[start:i+1]
                start = i+1
        if start < chunkend:
            yield l[start:chunkend]
        start = chunkend

def match_names(names, include_unmatched=False):
    strnames = [str(n) for n in names]
    result_chunks = (call_api(chunk)['items'] for chunk in chunklist(strnames))
    matched = (_make_taxon(d) for d in chain(*result_chunks)) 
    if include_unmatched:
        return zip(names, matched)
    else:
        return ((n,m) for n,m in zip(names, matched) if m is not None)

def streaming_match_taxa(taxa, taxa2=None, tracker=None):
    """Match taxa to the names in TNRS.
    
    If taxa is a single-use iterable (like an iterator over a file), taxa2 must
    be specified as a second equivalent iterable, which will be used up first.
    With a collection like a TaxonSet, taxa can be used twice.
    """
    tracker = prepare_tracker(tracker)
    if taxa2 is None:
        taxa2 = taxa
    
    names = list({t.name for t in taxa2})
    namematches = dict(match_names(names))
    del names, taxa2
    
    for taxon in taxa:
        tracker.reset()
        tracker.start_taxon(taxon)
        try:
            t2 = namematches[taxon.name]
        except KeyError:
            tracker.name_event(taxon.name, "no match")
            continue
        else:
            tracker.name_transform(taxon.name, t2.name, "matched")
            yield taxon, t2

def run_match_taxa(taxa, taxa2=None, **kwargs):
    for taxonpair in streaming_match_taxa(taxa, taxa2, **kwargs):
        pass
    
def match_taxa(taxa, taxa2=None, merge_info=combine_dicts, **kwargs):
    matching = streaming_match_taxa(taxa, taxa2, **kwargs)
    return build_matched_taxonset(matching, merge_info=merge_info)

def select(name, strict_authority=False, upgrade_subsp="nominal",
                    fuzzy=True, prefer_accepted='noauth', nameselector=None,
                                                        tracker=noop_tracker):
    if not isinstance(name, str):
        name = str(name)
    
    d = call_api([name])['items'][0]
    taxon = _make_taxon(d)
    if taxon is None:
        raise KeyError("No match found", name)
    return taxon



"""GUI Dialogs for loading CSV files.
"""
import os.path
import csv
from .qt import QtGui, QtCore

from . import iothread
from .ui.csvimport import Ui_CsvImportDialog
from .ui.csvimport_synonyms import Ui_CsvSynonymsImportDialog

supported_encodings = ["UTF-8", "Windows-1252", "IBM850", "GB18030"]

class CsvImportDialog(QtGui.QDialog):
    encoding = 'utf-8'
    UiClass = Ui_CsvImportDialog
    
    def __init__(self, filename, parent=None):
        super().__init__(parent=parent)
        self.ui = self.UiClass()
        self.ui.setupUi(self)
        
        self.ui.csv_encoding.addItems(supported_encodings)
        
        QtCore.QObject.connect(self.ui.csv_encoding,
            QtCore.SIGNAL("currentIndexChanged(QString)"), self.encoding_changed)
        
        self.filename = filename
        self.preview_file()
        
        ds_name = os.path.splitext(os.path.basename(filename))[0].replace("_"," ")
        self.ui.ds_name.setText(ds_name)
    
    def fill_column_dropdowns(self, fields):
        """Prepare CSV field dropdowns"""
        self.ui.namefield.clear()
        self.ui.authfield.clear()
        self.ui.namefield.addItems(fields)
        self.ui.authfield.addItems(fields)
        self.ui.authfield.setCurrentIndex(1)

    def preview_file(self):
        """Open file, grab column names, and preview the first 10 rows.
        
        preview should be a QTableWidget to put the preview in.
        """
        # Fill in preview table from CSV file
        with open(self.filename, encoding=self.encoding, errors='replace') as f:
            csvin = csv.reader(f)
            fields = next(csvin)
            head = []
            i = 0
            for row in csvin:
                head.append(row)
                i += 1
                if i >= 10:
                    break
        
        # Clear the old preview
        self.ui.preview.setColumnCount(0)
        self.ui.preview.setRowCount(0)
        
        for i in range(len(fields)):
            self.ui.preview.insertColumn(i)
        self.ui.preview.setHorizontalHeaderLabels(fields)
        for i, row in enumerate(head):
            self.ui.preview.insertRow(i)
            for j, cell in enumerate(row):
                self.ui.preview.setItem(i, j, QtGui.QTableWidgetItem(cell))
        
        self.fill_column_dropdowns(fields)
    
    def encoding_changed(self, new_encoding):
        "Switch encoding and update preview."
        self.encoding = new_encoding
        self.preview_file()

    def get_auth_field(self, prefix=''):
        """Get the location of the authorities in the format that the file_csv
        functions expect."""
        if getattr(self.ui, prefix+'noauth').isChecked():
            return None
        elif getattr(self.ui, prefix+'auth_with_name').isChecked():
            return True
        else:
            return getattr(self.ui, prefix+'authfield').currentText()

class CsvSynonymsImportDialog(CsvImportDialog):
    UiClass = Ui_CsvSynonymsImportDialog
    
    def fill_column_dropdowns(self, fields):
        self.ui.accnamefield.clear()
        self.ui.accauthfield.clear()
        self.ui.synnamefield.clear()
        self.ui.synauthfield.clear()
        self.ui.accnamefield.addItems(fields)
        self.ui.accauthfield.addItems(fields)
        self.ui.accauthfield.setCurrentIndex(1)
        self.ui.synnamefield.addItems(fields)
        self.ui.synnamefield.setCurrentIndex(2)
        self.ui.synauthfield.addItems(fields)
        self.ui.synauthfield.setCurrentIndex(3)

def load_csv(app):
    """Load a user-selected CSV file into the application."""
    filename = QtGui.QFileDialog.getOpenFileName(app, "Open file",
                            app.lastdir, "CSV (*.csv);;All files (*.*)")
    if not filename:
        return
    app.lastdir = os.path.dirname(filename)
    
    dialog = CsvImportDialog(filename, parent=app)
    
    if not dialog.exec_():
        return

    namefield = dialog.ui.namefield.currentText()
    authfield = dialog.get_auth_field()
    encoding = dialog.ui.csv_encoding.currentText()
    
    iothread.load_csv(app, dialog.ui.ds_name.text(), filename, encoding=encoding,
                                namefield=namefield, authfield=authfield)

def load_csv_synonyms(app):
    """Load a user-selected CSV file of synonyms into the application."""
    filename = QtGui.QFileDialog.getOpenFileName(app, "Open file",
                            app.lastdir, "CSV (*.csv);;All files (*.*)")
    if not filename:
        return
    app.lastdir = os.path.dirname(filename)
    
    dialog = CsvSynonymsImportDialog(filename, parent=app)
    
    if not dialog.exec_():
        return

    accnamefield = dialog.ui.accnamefield.currentText()
    synnamefield = dialog.ui.synnamefield.currentText()
    accauthfield = dialog.get_auth_field('acc')
    synauthfield = dialog.get_auth_field('syn')
    encoding = dialog.ui.csv_encoding.currentText()
    
    iothread.load_csv_synonyms(app, dialog.ui.ds_name.text(), filename, encoding,
                        accnamefield=accnamefield, accauthfield=accauthfield,
                        synnamefield=synnamefield, synauthfield=synauthfield)

def load_csv_individuals(app):
    """Load a user-selected CSV file of individual records into the application."""
    filename = QtGui.QFileDialog.getOpenFileName(app, "Open file",
                            app.lastdir, "CSV (*.csv);;All files (*.*)")
    if not filename:
        return
    app.lastdir = os.path.dirname(filename)
    
    dialog = CsvImportDialog(filename, parent=app)
    
    if not dialog.exec_():
        return

    namefield = dialog.ui.namefield.currentText()
    authfield = dialog.get_auth_field()
    encoding = dialog.ui.csv_encoding.currentText()
    
    iothread.load_csv_individuals(app, dialog.ui.ds_name.text(), filename, encoding=encoding,
                                namefield=namefield, authfield=authfield)



from distutils.core import setup
from os.path import join as pjoin

version = "0.14"

packages = ['taxonome',
            'taxonome.regions',
            'taxonome.regions.tests',
            'taxonome.external',
            'taxonome.services',
            'taxonome.taxa',
            'taxonome.taxa.tests',
            'taxonome.tests',
            'taxonome.utils',
            'taxonome.utils.tests',
            'taxonome.gui',
            'taxonome.gui.ui',
            ]
classifiers = ["Development Status :: 4 - Beta",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: MIT License",
               "Programming Language :: Python :: 3",
               "Topic :: Scientific/Engineering"
               ]
scripts = [pjoin("scripts", "taxonome"),
          ]

long_description = \
"""Taxonome is a set of tools for working with information on biological taxa.

In particular, it handles scientific names, and can match the names found in a
dataset to a given synonymy, taking into account spelling differences, synonyms
and homonyms.

It also includes functions to standardise species distribution information to the
regions defined by the `Taxonomic Diversity Working Group (TDWG)
<http://www.kew.org/science-research-data/kew-in-depth/gis/resources-and-publications/data/tdwg/index.htm>`_.

Taxonome includes a GUI interface for working with collections of taxa. This
requires `PyQt4 <http://pypi.python.org/pypi/PyQt>`_.
"""

setup(name="Taxonome",
      version=version,
      description="Tools for working with information on species or other taxa.",
      long_description=long_description,
      author="Thomas Kluyver",
      author_email="takowl@gmail.com",
      url="http://taxonome.bitbucket.org/",
      packages=packages,
      classifiers=classifiers,
      scripts=scripts,
      )
